from typing import Generator, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from ..flarum.core.notifications import Notifications
    from ..flarum.core.discussions import Discussions
    from ..flarum.core.posts import Posts
    from ..flarum.core.users import Users


import warnings


from . import ExtensionMixin
from ..session import FlarumUser

from ..flarum.core.filters import Filter
from ..flarum.core.discussions import Discussion


AUTHOR = 'skevo'
NAME = 'absolutely-all'
ID = f"{AUTHOR}-{NAME}"


SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class AbsolutelyAllFlarumUserMixin(FlarumUser):
    def absolutely_all_users(self, filter: Optional[Filter]=None)  -> Generator['Users', None, None]:
        """
            A generator that yields `Users` from entire forum, until there are `None` left. `Filter` compatible.
        """

        if isinstance(filter, Filter):
            _filter = filter

        else:
            _filter = Filter()


        _filter.page = 0
        users = self.get_users(_filter)


        while True:
            yield users

            users = self.get_users(_filter)
            _filter.page += 1

            if not users.next_link:
                break


    def absolutely_all_posts(self, filter: Optional[Filter]=None) -> Generator['Posts', None, None]:
        """
            A generator that yields `Posts` from entire forum, until there are `None` left. `Filter` compatible.
        """

        if isinstance(filter, Filter):
            _filter = filter

        else:
            _filter = Filter()


        _filter.page = 0
        posts = self.get_posts(_filter)


        while True:
            yield posts

            posts = self.get_posts(_filter)
            _filter.page += 1

            if not posts.next_link:
                break


    def absolutely_all_discussions(self, filter: Optional[Filter]=None) -> Generator['Discussions', None, None]:
        """
            A generator that yields `Discussions` from entire forum, until there are `None` left. `Filter` compatible.
        """

        if isinstance(filter, Filter):
            _filter = filter

        else:
            _filter = Filter()


        _filter.page = 0
        discussions = self.get_discussions(_filter)


        while True:
            yield discussions

            discussions = self.get_discussions(_filter)
            _filter.page += 1

            if not discussions.next_link:
                break


    def absolutely_all_notifications(self, filter: Optional[Filter]=None) -> Generator['Notifications', None, None]:
        """
            A generator that yields all of your `Notifications`, until there are `None` left. `Filter` compatible.
        """

        if isinstance(filter, Filter):
            _filter = filter

        else:
            _filter = Filter()


        _filter.page = 0
        discussions = self.get_notifications(_filter)


        while True:
            yield discussions

            discussions = self.get_notifications(_filter)
            _filter.page += 1

            if not discussions.next_link:
                break
    

    def get_all_posts_from_discussion(self, discussion: Discussion, at_once: int=50, force: bool=False) -> Generator['Posts', None, None]:
        """
            This fetches all posts from a long discussion where only post IDs are present.

            First, a list of all IDs is created from the API response. Then, IDs are broken into chunks of size `at_once` and
            yielded as `Posts`.

            Use `force=True` to bypass `at_once` being capped at 50, if more than 50.
        """

        if not isinstance(discussion, Discussion):
            raise TypeError("`discussion` parameter must be an instance of `Discussion`.")


        if at_once > 50 and not force:
            at_once = 50
            warnings.warn("`at_once` was capped at 50, because Flarum (by default/currently) doesn't support fetching more than 50 entires at once from API. Use `force=True` to bypass.")


        raw_posts = discussion.relationships.get("posts", {}).get("data", []) # type: list[dict]
        post_ids = [] # type: list[int]


        for raw_post in raw_posts:
            post_id = raw_post.get("id", None) # type: Optional[int]

            if post_id:
                post_ids.append(post_id)


        for id in range(0, len(post_ids), at_once): 
            yield self.get_posts(filter=Filter(limit=50, ids=post_ids[id:id + at_once]))



class AbsolutelyAllExtension(ExtensionMixin):
    """
        A pyFlarum extension. Allows you to fetch all specific data from a forum (e. g.: all discussions, all posts, etc.), until there are none left.

        Based on `Generator`, that yields in a while loop, until no `next_link` is present in the API.
    """

    def __init__(self):
        self.name = NAME
        self.author = AUTHOR
        self.id = ID


    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, FlarumUser, AbsolutelyAllFlarumUserMixin)



from typing import Literal, Optional, Iterable
import warnings


__all__ = ['Filter']



class Filter:
    """
        Represents a Flarum API filter as a `dict`.

        It allows you to filter discussions without having to manually specify URL parameters.
    """

    def __init__(self, order_by: Optional[Literal['commentCount', '-commentCount', 'createdAt', '-createdAt']]=None, query: Optional[str]=None, ids: Optional[Iterable['str | int']]=None, limit: int=20, page: Optional[int]=None, include: Optional[list[str]]=None, additional_data: Optional[dict]=None):
        """
            - `order_by` - gets passed into `?sort=` parameter. Common values are `commentCount`, `createdAt` and their reverse/negated values (prefixed with `-`)
            - `query` - the search query, passed into `?filter[q]=`. This can be a string. Flarum search bar uses this. Gambits such as `author:username` are supported.
            - `ids` - fetches entries with specific ids, passed into `?filter[id]=`. This is a list, that is then converted into comma separated string.
            - `limit` - limit of entires to fetch. Flarum (by default) allows a max. of 50 entries to be fetched at once. Passed into `?page[limit]=`
            - `page` - fetch a specific page of entires. This is actually an offset - which is determined by multiplying `page` with `limit` (see above).
            - `include` - include specific entries. See [included data](https://cwkevo.github.io/pyflarum/docs/#included-data). You will likely never use this.
            - `additional_data` - this is a `dict` (`parameter: value`) of additional search parameters that you might want to use. This can be used to overwrite previous filters.
        """

        if limit > 50:
            warnings.warn(f"Maximum amount of entries fetchable by 1 bulk API request for Flarum is 50 by default, you specified {limit}. This is a Flarum limitation to prevent flooding, and I can't bypass it. Please, lower the value to be 50 or below.")


        self.order_by = order_by
        self.query = query
        self.ids = ids
        self.limit = limit
        self.page = page
        self.include = include
        self.additional_data = additional_data

        self.as_json = self.to_dict


    @property
    def to_dict(self) -> dict:
        """
            Converts the filter to a `dict`, so that
            it can be sent to the API (`requests` module, see ["Passing parameters in URLs"](https://docs.python-requests.org/en/master/user/quickstart/#passing-parameters-in-urls)).

            An extension might add additional filter data after the filter was initialized
            (for example: `absolutely_all` needs to update page number to continuously yield results).
        """

        __data = {}

        if self.order_by:
            __data["sort"] = self.order_by

        if self.query:
            __data["filter[q]"] = self.query
        
        if self.ids:
            __data["filter[id]"] = ','.join(map(str, self.ids))

        if self.limit:
            __data["page[limit]"] = self.limit

        if self.page:
            __data["page[offset]"] = self.page * self.limit

        if self.include:
            __data["include"] = ','.join(self.include)


        if self.additional_data:
            __data.update(self.additional_data)
        
        return __data



from dotenv import load_dotenv
load_dotenv()

import os
import time

from pyflarum import FlarumUser

from pyflarum.extensions import absolutely_all
from pyflarum.extensions.flarum import Flarum_Approval


EXTENSIONS = [
    absolutely_all.AbsolutelyAllExtension,
    Flarum_Approval.ApprovalExtension
]


user = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'], extensions=EXTENSIONS) # type: absolutely_all.AbsolutelyAllFlarumUserMixin


if __name__ == "__main__":
    for discussions in user.absolutely_all_discussions():
        for discussion in discussions:
            time.sleep(5) # prevent 429
            full_discussion = discussion.get_full_data()

            for post in full_discussion.get_posts():
                post: Flarum_Approval.ApprovalPostFromNotificationMixin

                if not post.isApproved:
                    post.approve()
                    print(f"Approved {post.url}")

                else:
                    print(f"{post.url} is already approved.")



from typing import Optional

from datetime import datetime, timedelta

from .. import ExtensionMixin

from ...flarum.core.users import User, UserFromBulk
from ...error_handler import parse_request
from ...datetime_conversions import datetime_to_flarum, flarum_to_datetime


AUTHOR = 'flarum'
NAME = 'suspend'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class SuspendUserMixin(UserFromBulk):
    @property
    def canSuspend(self) -> bool:
        return self.attributes.get("canSuspend", False)


    @property
    def suspendedUntil(self):
        raw = self.attributes.get("suspendedUntil", None)

        return flarum_to_datetime(raw)


    def suspend(self, suspended_until: Optional[datetime]=None, suspended_for: Optional[timedelta]=None) -> User:
        if suspended_until:
            until = suspended_until

        else:
            if suspended_for:
                raw_until = datetime.now() + suspended_for
                until = raw_until

            else:
                until = datetime(6669, 2, 6)


        post_data = {
            "data": {
                "type": "users",
                "id": "3",
                "attributes": {
                    "suspendedUntil": datetime_to_flarum(until)
                }
            }
        }

        raw = self.user.session.patch(f"{self.user.api_urls['users']}/{self.id}", json=post_data)
        json = parse_request(raw)

        return User(user=self.user, _fetched_data=json)



class SuspendExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, UserFromBulk, SuspendUserMixin)



from typing import Iterator, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from ...session import FlarumUser

from datetime import datetime

from ..core import BaseFlarumBulkObject, BaseFlarumIndividualObject
from ..core.discussions import DiscussionFromNotification
from ..core.users import UserFromNotification
from ..core.posts import PostFromNotification

from ...error_handler import parse_request
from ...datetime_conversions import flarum_to_datetime


__all__ = ['Notifications', 'Notification']



class Notifications(BaseFlarumBulkObject):
    """
        A data of multiple notifications fetched from the API.
    """


    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        return super().__init__(user=user, _fetched_data=_fetched_data, _listclass=Notification, _required_type='notifications')


    def mark_all_as_read(self) -> True:
        """
            Marks all notifications as read. Returns `True` when successful.
        """

        return self.user.mark_all_notifications_as_read()


    if TYPE_CHECKING:
        def __getitem__(self, key: int) -> 'Notification': ...
        def __iter__(self) -> Iterator['Notification']: ...



class Notification(BaseFlarumIndividualObject):
    """
        Notification.
    """


    @property
    def contentType(self) -> Optional[str]:
        """
            The content type of the notification.

            Examples: `newPost`, `postLiked`, etc...
        """

        return self.attributes.get("contentType", None)


    @property
    def content(self) -> Optional[dict]:
        """
            The `dict` of the notification's content.
        """

        return self.attributes.get("content", None)


    @property
    def new_post_number(self) -> Optional[int]:
        """
            The new number of the potential post that triggered
            the notification.
        """

        if self.content and self.contentType == "newPost":
            post_number = self.content.get("postNumber", None)

            if post_number:
                return int(post_number)


    @property
    def reply_number(self) -> Optional[int]:
        """
            The number of the reply post that possibly triggered
            the notification.
        """

        if self.content and self.contentType == "postMentioned":
            reply_number = self.content.get("replyNumber", None)

            if reply_number:
                return int(reply_number)


    @property
    def createdAt(self) -> Optional[datetime]:
        """
            The `datetime` of when was this notification triggered/created at.
        """

        raw = self.attributes.get("createdAt", None)

        return flarum_to_datetime(raw)


    @property
    def isRead(self) -> bool:
        """
            Whether or not the notification was read by you.
        """

        return self.attributes.get("isRead", False)


    def from_user(self) -> Optional[UserFromNotification]:
        """
            From which user does the notification originate from?

            Returns `pyflarum.flarum.core.users.UserFromNotification`.
        """

        id = self.relationships.get("fromUser", {}).get("data", {}).get("id", None)
        
        for raw_user in self._parent_included:
            if raw_user.get("id", None) == id and raw_user.get("type", None) == 'users':
                user = UserFromNotification(user=self.user, _fetched_data=dict(data=raw_user))
                return user

        return None


    def get_subject(self) -> Optional['DiscussionFromNotification | PostFromNotification']:
        """
            Returns the subject of the notification, either one of these:
            - `pyflarum.flarum.core.discussions.DiscussionFromNotification`
            - `pyflarum.flarum.core.posts.PostFromNotification`
        """

        id = self.relationships.get("subject", {}).get("data", {}).get("id", None)
        
        for raw_subject in self._parent_included:
            if raw_subject.get("id", None) == id:
                notification_type = raw_subject.get("type", None)

                if notification_type == 'discussions':
                    return DiscussionFromNotification(user=self.user, _fetched_data=dict(data=raw_subject))

                elif notification_type == 'posts':
                    return PostFromNotification(user=self.user, _fetched_data=dict(data=raw_subject, _parent_included=self._parent_included))

        return None


    def mark_as_read(self) -> True:
        """
            Marks the notification as read.

            Returns `True` when successful.
        """

        post_data = {"is_read": True}
        raw = self.user.session.patch(f"{self.user.api_urls['notifications']}/{self.id}", json=post_data)
        parse_request(raw)

        return True



from dotenv import load_dotenv
load_dotenv()

import os
from datetime import datetime

from pyflarum import FlarumUser
from pyflarum.flarum.core.posts import PreparedPost


user = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'])


if __name__ == "__main__":
    # Edit first post of discussion with ID 16 to the current time.
    first_discussion = user.get_discussions()[0].get_full_data()
    first_post = first_discussion.get_posts()[0]

    if first_post.contentType == 'comment':
        edited_content = f"""{first_post.content}\n###### BTW, The current time for me, the bot, is: {datetime.now().time().strftime(r"%H hours, %M minutes and %S seconds.")}."""
        edit = PreparedPost(user=user, discussion=first_discussion, content=edited_content)

        edited = first_post.edit(edit)
        print(f"Successfuly edited post {edited.id} - {edited.url}")

    else:
        print(f"First post of discussion ID {first_discussion.id} is not a comment (but {first_post.contentType}), editing that is not possible.")



"""
    Custom types for pyFlarum.

    ### Types:
    - `AnyUser` - `User` | `UserFromBulk` | `UserFromNotification`
    - `AnyDiscussion` - `Discussion` | `DiscussionFromBulk` | `DiscussionFromNotification`
    - `AnyPost` - `Post` | `PostFromBulk` | `PostFromNotification`
"""

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .session import FlarumUser

    from .flarum.core import BaseFlarumIndividualObject

    from .flarum.core.users import User, UserFromBulk, UserFromNotification
    from .flarum.core.discussions import Discussion, DiscussionFromBulk, DiscussionFromNotification
    from .flarum.core.posts import Post, PostFromBulk, PostFromNotification, PostFromDiscussion

    AnyUser: 'User | UserFromBulk | UserFromNotification'
    AnyDiscussion: 'Discussion | DiscussionFromBulk | DiscussionFromNotification'
    AnyPost:  'Post | PostFromBulk | PostFromNotification | PostFromDiscussion'

    class AnyFlarumClass(BaseFlarumIndividualObject):
        def __init__(self, user: 'FlarumUser', _fetched_data: dict) -> None: ...
        def __call__(self, user: 'FlarumUser', _fetched_data: dict): ...



from .. import ExtensionMixin

from ...flarum.core.discussions import DiscussionFromNotification
from ...flarum.core.posts import Post, PostFromNotification

from ...error_handler import FlarumError, parse_request


AUTHOR = 'flarum'
NAME = 'approval'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class ApprovalDiscussionFromNotificationMixin(DiscussionFromNotification):
    @property
    def isApproved(self) -> bool:
        """
            Whether or not the discussion is approved.
        """

        return self.attributes.get("isApproved", False)



class ApprovalPostFromNotificationMixin(PostFromNotification):
    @property
    def isApproved(self) -> bool:
        """
            Whether or not the post is approved.
        """

        return self.attributes.get("isApproved", False)


    @property
    def canApprove(self) -> bool:
        """
            Whether or not you are able to approve the post
        """

        return self.attributes.get("canApprove", False)


    def approve(self, force: bool=False) -> Post:
        """
            Approve the post. Use `force` to approve despite the post being approved already, and do not raise `FlarumError`.
        """

        if self.isApproved and not force:
            raise FlarumError(f'Post ID {self.id} is already approved. Use `force = True` to bypass this error.')


        patch_data = {
            "data": {
                "type": "posts",
                "id": self.id,
                "attributes": {
                    "isApproved": True
                }
            }
        }

        raw = self.user.session.patch(f"{self.user.api_urls['posts']}/{self.id}", json=patch_data)
        json = parse_request(raw)

        return Post(user=self.user, _fetched_data=json)



class ApprovalExtension(ExtensionMixin):
    """
        https://packagist.org/packages/flarum/approval
    """

    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, DiscussionFromNotification, ApprovalDiscussionFromNotificationMixin)
        super().mixin(self, PostFromNotification, ApprovalPostFromNotificationMixin)



from dotenv import load_dotenv
load_dotenv()

import os

from pyflarum import FlarumUser

from pyflarum.extensions import absolutely_all
from pyflarum.extensions.flarum import Flarum_Lock


EXTENSIONS = [
    absolutely_all.AbsolutelyAllExtension,
    Flarum_Lock.LockExtension
]


USER = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'], extensions=EXTENSIONS) # type: absolutely_all.AbsolutelyAllFlarumUserMixin


if __name__ == "__main__":
    for discussions in USER.absolutely_all_discussions():
        for discussion in discussions:
            discussion: Flarum_Lock.LockDiscussionFromBulkMixin

            if not discussion.isLocked:
                locked = discussion.lock()
                print(f"Locked {locked.url}")

            else:
                print(f"{discussion.url} is already locked.")



from dotenv import load_dotenv
load_dotenv()

import os
import requests

from pyflarum import FlarumUser

from pyflarum.flarum.core.notifications import Notification
from pyflarum.flarum.core.posts import PostFromNotification, PreparedPost

from pyflarum.extensions import watch
from pyflarum.extensions import commands
from pyflarum.extensions.flarum import Flarum_Likes


EXTENSIONS = [
    watch.WatchExtension,
    commands.CommandsExtension,
    Flarum_Likes.LikesExtension
]

user = FlarumUser(forum_url=os.environ['forum_url'], username_or_email="test", password=os.environ['account_password'], extensions=EXTENSIONS) # type: watch.WatchFlarumUserMixin | commands.CommandsFlarumUserMixin
api_key = os.environ['openweather_api_key']

weather_post = """**{city}**, `{country}` has **{weather}** and __{temperature} °C__"""

def on_notification(notification: Notification):
    subject = notification.get_subject()

    if isinstance(subject, PostFromNotification):
        subject: Flarum_Likes.LikesPostFromNotificationMixin

        if user.is_mentioned_in(subject.content):
            print("New mention!")
            command = user.parse_as_command(subject.content)

            if command[0].lower() == 'weather':
                if command[1]:
                    city = command[1].strip()
                    weather_data = requests.get(f"https://api.openweathermap.org/data/2.5/weather?q={city}&appid={api_key}&units=metric").json() # type: dict

                    reply_post = PreparedPost(user=user, content=weather_post.format(
                        city=city,
                        country=weather_data.get("sys", {}).get("country", "N/A"),
                        weather=weather_data.get("weather", [{}])[0].get("description", "unknown weather"),
                        temperature=weather_data.get("main", {}).get("temp", "N/A"))
                    )
        
                    subject.reply_to(reply_post)

                    print(f"Replied to post in discussion ID {subject.id} (weather)")

                else:
                    subject.reply_to(PreparedPost(user=user, content="Please, specify the city."))

                    print("No city specified for 'weather' command")
            
            elif command[0].lower() == "like":
                liked = subject.like()
                print(f"Successfuly liked post {liked.id} ({liked.url})")

            else:
                print("Match not found for command data:", command)
            
            notification.mark_as_read()


if __name__ == "__main__":
    print(f"Watching notifications as user '{user.username}'...")
    user.watch_notifications(on_notification, auto_mark_as_read=False)



from typing import Optional


from ..core import BaseFlarumIndividualObject
from ..core.groups import Group


__all__ = ['Forum']


class Forum(BaseFlarumIndividualObject):
    """
        Entire forum, lives under the main `/api` route.
    """


    @property
    def title(self) -> Optional[str]:
        """
            The forum's title.
        """

        return self.attributes.get("title", None)


    @property
    def description(self) -> Optional[str]:
        """
            The description of the forum.
        """

        return self.attributes.get("description", None)


    @property
    def showLanguageSelector(self) -> bool:
        """
            Whether or not the language selector is available.
        """

        return self.attributes.get("showLanguageSelector", False)


    @property
    def baseUrl(self) -> Optional[str]:
        """
            Base URL of the forum/where the forum is located at.
        """

        return self.attributes.get("baseUrl", None)


    @property
    def basePath(self) -> Optional[str]:
        """
            Base path to the forum.
        """

        return self.attributes.get("basePath", None)


    @property
    def debug(self) -> bool:
        """
            Whether or not debug mode is enabled.
        """

        return self.attributes.get("debug", False)


    @property
    def apiUrl(self) -> Optional[str]:
        """
            The API URL of the forum.
        """

        return self.attributes.get("apiUrl", None)


    @property
    def welcomeTitle(self) -> Optional[str]:
        """
            The title of the welcome message box of the forum.
        """

        return self.attributes.get("welcomeTitle", None)


    @property
    def welcomeMessage(self) -> Optional[str]:
        """
            The welcome message of the forum (shown in the welcome box).
        """

        return self.attributes.get("welcomeMessage", None)


    @property
    def themePrimaryColor(self) -> Optional[str]:
        """
            Forum's primary color in HEX format.
        """

        return self.attributes.get("themePrimaryColor", None)


    @property
    def themeSecondaryColor(self) -> Optional[str]:
        """
            Forum's secondary color in HEX format.
        """

        return self.attributes.get("themeSecondaryColor", None)


    @property
    def logoUrl(self) -> Optional[str]:
        """
            URL to forum's logo.
        """

        return self.attributes.get("logoUrl", None)


    @property
    def faviconUrl(self) -> Optional[str]:
        """
            URL to forum's favicon.
        """

        return self.attributes.get("faviconUrl", None)


    @property
    def headerHtml(self) -> Optional[str]:
        """
            The header HTML of the forum.
        """

        return self.attributes.get("headerHtml", None)


    @property
    def footerHtml(self) -> Optional[str]:
        """
            The footer HTML of the forum.
        """

        return self.attributes.get("footerHtml", None)


    @property
    def allowSignUp(self) -> bool:
        """
            Whether or not signup is allowed.
        """

        return self.attributes.get("allowSignUp", False)


    @property
    def defaultRoute(self) -> Optional[str]:
        """
            The homepage of the forum (default route)
        """

        return self.attributes.get("defaultRoute", None)


    @property
    def canViewForum(self) -> bool:
        """
            Whether or not you are allowed to view the forum.
        """

        return self.attributes.get("canViewForum", False)


    @property
    def canStartDiscussion(self) -> bool:
        """
            Whether or not you are allowed to start a discussion.
        """

        return self.attributes.get("canStartDiscussion", False)


    @property
    def canSearchUsers(self) -> bool:
        """
            Whether or not you are able to search for users.
        """

        return self.attributes.get("canSearchUsers", False)


    @property
    def adminUrl(self) -> Optional[str]:
        """
            The administration panel URL of the forum.
        """

        return self.attributes.get("adminUrl", None)


    @property
    def version(self) -> Optional[str]:
        """
            The Flarum version this forum is running on.
        """

        return self.attributes.get("version", None)


    @property
    def allowUsernameMentionFormat(self) -> bool:
        return self.attributes.get("allowUsernameMentionFormat", False)


    def get_groups(self) -> list[Group]:
        """
            Obtains the forum groups.

            Returns a list of `Group` objects.
        """

        all_groups = list()

        for raw_group in self.relationships.get("groups", {}).get("data", [{}]):
            if raw_group.get("type", None) == 'groups':
                group = Group(user=self.user, _fetched_data=dict(data=raw_group))

                all_groups.append(group)

        return all_groups


    # Required: `Forum` is not a standard bulk route, so we can't inherit from that.
    @property
    def included(self) -> list[dict]:
        """
            Returns raw list of JSON included data.

            Learn more about included data [here](https://cwkevo.github.io/pyflarum/docs/#included-data)
        """

        return self.get("included", [{}])



from typing import Iterator, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from ...session import FlarumUser

from datetime import datetime

from ..core import BaseFlarumBulkObject, BaseFlarumIndividualObject
from ...datetime_conversions import flarum_to_datetime


__all__ = ['Users', 'User', 'UserFromBulk', 'UserFromNotification', 'MyUser']



class Users(BaseFlarumBulkObject):
    """
        A data of multiple users fetched from the API.
    """


    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        return super().__init__(user=user, _fetched_data=_fetched_data, _listclass=UserFromBulk, _required_type='users')


    if TYPE_CHECKING:
        def __getitem__(self, key: int) -> 'UserFromBulk': ...
        def __iter__(self) -> Iterator['UserFromBulk']: ...



class UserFromNotification(BaseFlarumIndividualObject):
    """
        An user from `BaseNotification`
    """


    @property
    def username(self) -> Optional[str]:
        """
            The user's username.
        """

        return self.attributes.get("username", None)


    @property
    def email(self) -> Optional[str]:
        """
            The user's E-mail, if you have permission to view it.
        """

        return self.attributes.get("email", None)


    @property
    def isEmailConfirmed(self) -> bool:
        """
            Whether or not this user confirmed their E-mail address.

            You must have the permission to view the user's E-mail address
            in order to know this too in the first place.
        """

        return self.attributes.get("isEmailConfirmed", False)


    @property
    def displayName(self) -> Optional[str]:
        """
            The display name/nickname of the user.
        """

        return self.attributes.get("displayName", None)


    @property
    def avatarUrl(self) -> Optional[str]:
        """
            The user's avatar URL.
        """

        return self.attributes.get("avatarUrl", None)


    @property
    def slug(self) -> Optional[str]:
        """
            The user's slug.
        """

        return self.attributes.get("slug", None)



class UserFromBulk(UserFromNotification):
    """
        An user from `Users`.
    """


    @property
    def joinTime(self) -> Optional[datetime]:
        """
            The `datetime` of when the user had joined this forum.
        """

        raw = self.attributes.get("joinTime", None)

        return flarum_to_datetime(raw)


    @property
    def discussionCount(self) -> Optional[int]:
        """
            The user's discussion count.
        """

        return self.attributes.get("discussionCount", None)


    @property
    def commentCount(self) -> Optional[int]:
        """
            The user's comment/post count.
        """

        return self.attributes.get("commentCount", None)


    @property
    def canEdit(self) -> bool:
        """
            Whether or not you are able to edit this user.
        """

        return self.attributes.get("canEdit", False)


    @property
    def canEditCredentials(self) -> bool:
        """
            Whether or not you are able to edit this user's credentials.
        """

        return self.attributes.get("canEditCredentials", False)


    @property
    def canEditGroups(self) -> bool:
        """
            Whether or not you are able to edit this user's groups.
        """

        return self.attributes.get("canEditGroups", False)


    @property
    def canDelete(self) -> bool:
        """
            Whether or not you are able to scronch this user forever.
        """

        return self.attributes.get("canDelete", False)



class User(UserFromBulk):
    """
        An user that was fetched from the API.
    """

    pass



class MyUser(User):
    """
        Your user, contains fullest user data.
    """

    @property
    def markedAllAsReadAt(self) -> Optional[datetime]:
        """
            When did you mark all discussions as read.
        """

        raw = self.attributes.get("markedAllAsReadAt", None)

        return flarum_to_datetime(raw)
    

    @property
    def unreadNotificationCount(self) -> Optional[int]:
        """
            Amount of your unread notifications.
        """

        return self.attributes.get("unreadNotificationCount", None)


    @property
    def newNotificationCount(self) -> Optional[int]:
        """
            Amount of your new notifications.
        """

        return self.attributes.get("newNotificationCount", None)


    @property
    def preferences(self) -> dict:
        """
            A raw `dict` of your preferences (for notifications).
        """

        return self.attributes.get("preferences", {})




from typing import Any, Iterator, Literal, NoReturn, TYPE_CHECKING, Optional

# Avoid my greatest enemy in Python: circular import:
if TYPE_CHECKING:
    from ...session import FlarumUser

from datetime import datetime

from ..core import BaseFlarumBulkObject, BaseFlarumIndividualObject
from ..core.users import UserFromBulk, UserFromNotification
from ...error_handler import FlarumError, parse_request
from ...datetime_conversions import flarum_to_datetime


__all__ = ['PreparedDiscussion', 'Discussions', 'Discussion', 'DiscussionFromBulk', 'DiscussionFromNotification']



class PreparedDiscussion(BaseFlarumIndividualObject):
    """
        A prepared discussion that can be sent to the API.
    """

    def __init__(self, user: 'FlarumUser', title: Optional[str]=None, content: Optional[str]=None):
        """
            ### Parameters:
            - `user` - the `pyflarum.session.FlarumUser` object that will create the discussion
            (see `PreparedDiscussion.post()`).
            - `title` - the discussion's title.
            - `content` - discussion's content. You can use the traditional Flarum's markdown syntax.
        """

        self.user = user
        self.title = title
        self.content = content

        self.as_json = self.to_dict
        super().__init__(user=self.user, _fetched_data=self.as_json)


    @property
    def to_dict(self) -> dict:
        """
            Converts the discussion to a `dict`, so that
            it can be sent to the API.

            An extension might add additional data during runtime. This is the
            most basic template that Flarum requires when creating a discussion.
        """

        data = {
            "data": {
                "type": "discussions",
                "attributes": {
                    "title": self.title,
                    "content": self.content
                }
            }
        }

        return data


    def post(self) -> 'Discussion':
        """
            Posts/creates the discussion. Raises `FlarumError` if it failed, otherwise the new `Discussion` is returned.
            This is the same as `PreparedDiscussion.create()`.
        """

        if not isinstance(self.title, str) or not isinstance(self.content, str):
            raise TypeError(f"Both `title` and `content` parameters must be a `str`.")

        raw = self.user.session.post(self.user.api_urls['discussions'], json=self.as_json)
        json = parse_request(raw)

        return Discussion(user=self.user, _fetched_data=json)
    create = post



class Discussions(BaseFlarumBulkObject):
    """
        A data of multiple discussions fetched from `/api/discussions`.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        return super().__init__(user=user, _fetched_data=_fetched_data, _listclass=DiscussionFromBulk, _required_type='discussions')


    if TYPE_CHECKING:
        def __getitem__(self, key: int) -> 'DiscussionFromBulk': ...
        def __iter__(self) -> Iterator['DiscussionFromBulk']: ...



class DiscussionFromNotification(BaseFlarumIndividualObject):
    """
        A discussion from `Notification`. Contains the least
        data from all of the discussion classes (see [inheritance](https://cwkevo.github.io/pyflarum/docs/#class-inheritance)).
    """


    def get_full_data(self) -> 'Discussion':
        """
            Makes an additional API call to fetch the full data of the discussion, e. g.
            the top-level discussion class (`Discussion`).

            Learn more about [inheritance](https://cwkevo.github.io/pyflarum/docs/#class-inheritance).
        """

        raw = self.user.session.get(f"{self.user.api_urls['discussions']}/{self.id}")
        json = parse_request(raw)

        return Discussion(user=self.user, _fetched_data=json)


    @property
    def title(self) -> Optional[str]:
        """
            Returns the discussion's title.
        """

        return self.attributes.get("title", None)


    @property
    def slug(self) -> Optional[str]:
        """
            Returns the discussion's slug
            (consists of ID and dash separated words from discussion's title,
            e. g. `123-some-title`).
        """

        return self.attributes.get("slug", None)


    def __restore_or_hide(self, hide: bool) -> 'Discussion':
        """
            Either restores or hides the discussion.

            This function was made to prevent code repetition - please
            use `Discussion.hide()` and `Discussion.restore()` instead.
        """

        patch_data = {
            "data": {
                "type": "discussions",
                "id": self.id,
                "attributes": {
                    "isHidden": hide
                }
            }
        }

        raw = self.user.session.patch(f"{self.user.api_urls['discussions']}/{self.id}", json=patch_data)
        json = parse_request(raw)

        return Discussion(user=self.user, _fetched_data=json)


    def hide(self) -> 'Discussion':
        """
            Hides the discussion.
            Raises `FlarumError` if it failed, otherwise the new discussion is returned.
        """

        return self.__restore_or_hide(hide=True)


    def restore(self) -> 'Discussion':
        """
            Restores the discussion (unhides).
            Raises `FlarumError` if it failed, otherwise the new discussion is returned.

            `Discussion.unhide()` does the same thing.
        """

        return self.__restore_or_hide(hide=False)
    unhide = restore


    def delete(self) -> 'Discussion':
        """
            Deletes a discussion forever - this action is irreversible!
            Returns `True` on success, `FlarumError` otherwise.
        """

        raw = self.user.session.delete(f"{self.user.api_urls['discussions']}/{self.id}")
        parse_request(raw)

        return True



# Circular:
from ...flarum.core.posts import PostFromDiscussion
class DiscussionFromBulk(DiscussionFromNotification):
    """
        A discussion from `Discussions`.
    """


    @property
    def url(self):
        """
            Returns the discussion's URL (including slug, if it's available).
        """

        slug = self.slug

        if slug:
            return f"{self.user.forum_url}/d/{slug}"

        else:
            return f"{self.user.forum_url}/d/{self.id}"


    @property
    def commentCount(self) -> Optional[str]:
        """
            Obtains the comment count of the discussion.

            A comment is a post made by an user.
        """

        return self.attributes.get("commentCount", None)


    @property
    def participantCount(self) -> Optional[str]:
        """
            The participant count of the discussion. This is
            the number of all users that have participated in a
            discussion by posting.
        """

        return self.attributes.get("participantCount", None)


    @property
    def createdAt(self) -> Optional[datetime]:
        """
            The `datetime` of when this discussion was created at.
        """

        raw = self.attributes.get("createdAt", None)

        return flarum_to_datetime(raw)


    @property
    def lastPostedAt(self) -> Optional[datetime]:
        """
            The `datetime` of when the last post in this
            discussion was made, e. g. when was this
            discussion last updated at.
        """

        raw = self.attributes.get("lastPostedAt", None)

        return flarum_to_datetime(raw)


    @property
    def lastPostNumber(self) -> Optional[int]:
        """
            Returns the number of the newest post in the
            discussion.
        """

        return self.attributes.get("lastPostNumber", None)


    @property
    def lastReadPostNumber(self) -> Optional[int]:
        """
            Number of a post that you've last read in the discussion.
        """

        return self.attributes.get("lastReadPostNumber", None)


    @property
    def canReply(self) -> bool:
        """
            Whether or not you are able to create a post in the discussion.
        """

        return self.attributes.get("canReply", False)


    @property
    def canRename(self) -> bool:
        """
            Whether or not you are able to rename the discussion.
        """

        return self.attributes.get("canRename", False)


    @property
    def canDelete(self) -> bool:
        """
            Whether or not you are able to delete the discussion.
        """

        return self.attributes.get("canDelete", False)


    @property
    def canHide(self) -> bool:
        """
            Whether or not you are able to hide the discussion.
        """

        return self.attributes.get("canHide", False)


    @property
    def lastReadAt(self) -> Optional[datetime]:
        """
            The `datetime` when you last read that discussion.
        """

        raw = self.attributes.get("lastReadAt", None)

        return flarum_to_datetime(raw)


    @property
    def isHidden(self) -> bool:
        """
            Whether or not the discussion is hidden. This happens when
            you delete the discussion for the first time.
        """

        return self.attributes.get("isHidden", False)


    def get_author(self) -> Optional[UserFromNotification]:
        """
            Obtains the author of the discussion.

            It returns `pyflarum.flarum.core.users.UserFromNotification` because it's JSON
            data matches the data of user from notification. If no user is found, `None` is
            returned.

            This works by fetching it from the `_parent_included` property.
        """

        id = self.relationships.get("user", {}).get("data", {}).get("id", None)

        for raw_user in self._parent_included:
            if raw_user.get("id", None) == id and raw_user.get("type", None) == 'users':
                user = UserFromNotification(user=self.user, _fetched_data=dict(data=raw_user))

                return user

        return None


    def get_last_posted_user(self) -> Optional[UserFromNotification]:
        """
            Obtains the user that posted the latest post in the discussion.

            It returns `pyflarum.flarum.core.users.UserFromNotification` because it's JSON
            data matches the data of user from notification. If no user is found, `None` is returned.

            This works by fetching it from the `_parent_included` property.
        """

        id = self.relationships.get("lastPostedUser", {}).get("data", {}).get("id", None)

        for raw_user in self._parent_included:
            if raw_user.get("id", None) == id and raw_user.get("type", None) == 'users':
                user = UserFromNotification(user=self.user, _fetched_data=dict(data=raw_user))

                return user

        return None


    def get_first_post(self) -> Optional[PostFromDiscussion]:
        """
            Obtains the first post of the discussion. If no post is found,
            `None` is returned.

            This works by fetching it from the `_parent_included` property.
        """

        id = self.relationships.get("firstPost", {}).get("data", {}).get("id", None)

        for raw_post in self._parent_included:
            if raw_post.get("id", None) == id and raw_post.get("type", None) == 'users':
                post = PostFromDiscussion(user=self.user, _fetched_data=dict(data=raw_post))

                return post

        return None


    def __restore_or_hide(self, hide: bool, force: bool=False):
        """
            Either restores or hides the post.

            This function was made to prevent repetitve code, and you should use
            `hide()` or `restore()` (`unhide()`) instead.
        """

        if hide:
            if self.isHidden and not force:
                raise FlarumError(f"Discussion {self.id} is already hidden. Use `force = True` to ignore this error.")

        else:
            if not self.isHidden and not force:
                raise FlarumError(f"Discussion {self.id} is already restored. Use `force = True` to ignore this error.")


        if not self.canHide and not force:
            raise FlarumError(f'You do not have permission to {"hide" if hide else "unhide"} this discussion ({self.id}). Use `force = True` to ignore this error.')

        return super().__restore_or_hide(hide=hide)


    def hide(self, force: bool=False):
        """
            Hides the discussion from the sight of other unprivileged users
            that are not worthy to view such thread.
        """

        return self.__restore_or_hide(hide=True, force=force)


    def restore(self, force: bool=False):
        """
            Restores the discussion (unhides it), bringing it back to life.
        """

        return self.__restore_or_hide(hide=False, force=force)
    unhide = restore


    def delete(self, force: bool=False) -> bool:
        """
            Scronches the discussion forever. This cannot be reverted.

            Use `force=True` to attempt to delete the discussion even if the API states that you can't.
        """

        if not self.canDelete and not force:
            raise FlarumError(f'You do not have permission to delete this discussion ({self.id})')

        else:
            return super().delete()


# Circular:
from ...flarum.core.posts import PostFromBulk
class Discussion(DiscussionFromBulk):
    """
        A Flarum discussion, obtained directly from the API by ID.
        
        This is the top-level discussion object that contains all the properties of a discussion, and
        inherits properties from all previous discussion-like objects.

        Learn more about inheritance [here](https://cwkevo.github.io/pyflarum/docs/#class-inheritance)
    """


    @property
    def included(self) -> list[dict]:
        """
            Returns raw list of JSON included data.

            Learn more about included data [here](https://cwkevo.github.io/pyflarum/docs/#included-data)
        """

        return self.get("included", [{}])


    def get_author(self, mode: 'Any | Literal["first_number"]'='first_number') -> UserFromBulk:
        """
            Obtains the discussion's author, AKA. the author
            of the post with number 1 in a discussion.

            `mode` allows you to specify the mode that is used to determine whether or not the post is the first post
            of the discussion.

            - `'first_number'` - checks if the number of the post is 1 - if yes, it fetches that post's author.
            - `Any` - if anything other than `'first_number'` is passed (e. g. `'first_user``, but this can be anything), then this
            returns the author of the first post in the JSON.
            I am not sure how reliable is this, and whether or not the posts are actually ordered correctly in the API, so it's
            probably a good idea to also check if the number of the post is 1 - but then again, what if the first post gets removed?
        """

        for post in self.get_posts():
            if mode == 'first_number':
                if post.number == 1:
                    author = post.get_author()

                    return author

            else:
                author = post.get_author()
                return author

        return None


    def get_posts(self) -> list[PostFromBulk]:
        """
            Returns a list of `pyflarum.flarum.core.posts.PostFromBulk` objects.

            It might seem strange why this doesn't return `pyflarum.flarum.core.posts.PostFromDiscussion` instead,
            but these posts are in fact identical to `pyflarum.flarum.core.posts.PostFromBulk`, that's why they are returned.

            `pyflarum.flarum.core.posts.PostFromDiscussion` comes from `pyflarum.flarum.core.discussions.DiscussionFromBulk` instead.
        """

        all_posts = list()
        raw_posts = self.relationships.get("posts", {}).get("data", [{}]) # type: list[dict]

        for raw_post in raw_posts:
            if raw_post.get("type", None) == 'posts':
                id_to_find = raw_post.get("id", None)

                if id_to_find:
                    for possible_raw_post in self.included:
                        if (possible_raw_post.get("type", None) == 'posts') and (possible_raw_post.get("id", None) == id_to_find):
                            post = PostFromBulk(user=self.user, _fetched_data=dict(data=possible_raw_post, _parent_included=self.included))

                            all_posts.append(post)

        return all_posts


    def get_first_post(self) -> NoReturn:
        """
            The `Discussion` object does not have the first post's JSON data in it's own JSON. Because of Python's subclass inheritance, this
            function was included in `Discussion`, but it does not work!

            ### Alternative:

            ```python
            discussion = user.get_discussion_by_id(1)
            first_post = discussion.get_posts()[0]
            ```
        """

        raise NotImplementedError("`Discussion` does not have the first post data to fetch. Please, use the snippet in this function's docstring instead to obtain the first post.")



from typing import Optional

from .. import ExtensionMixin
from ...flarum.core.discussions import DiscussionFromBulk


AUTHOR = 'fof'
NAME = 'prevent-necrobumping'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class PreventNecrobumpingDiscussionMixin(DiscussionFromBulk):
    @property
    def fof_prevent_necrobumping(self) -> Optional[int]:
        """
            I have no idea what this does either, sorry.
        """

        return self.attributes.get("fof-prevent-necrobumping", None)



class PreventNecrobumpingExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, DiscussionFromBulk, PreventNecrobumpingDiscussionMixin)



from dotenv import load_dotenv
load_dotenv()

import os

from pyflarum import FlarumUser
from pyflarum.flarum.core.discussions import PreparedDiscussion


user = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'])


if __name__ == "__main__":
    to_post = PreparedDiscussion(user=user, title='Test', content='testingy')
    posted = to_post.post()
    print(posted.url)



from typing import Optional

from . import ExtensionMixin
from ..session import FlarumUser

from ..flarum.core.users import User
from ..flarum.core.filters import Filter

from ..error_handler import parse_request


AUTHOR = 'skevo'
NAME = 'advanced-search'
ID = f"{AUTHOR}-{NAME}"


SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class AdvancedSearchFlarumUserMixin(FlarumUser):
    def get_user_by_username(self, username: str) -> Optional[User]:
        filter = Filter(query=username, limit=1)

        raw = self.session.get(f"{self.api_urls['users']}", params=filter.to_dict)
        json = parse_request(raw)

        for possible_user in json.get("data", [{}]):
            if possible_user.get("attributes", {}).get("username", None) == username:
                return User(user=self, _fetched_data=dict(data=possible_user))

        return None



class AdvancedSearchExtension(ExtensionMixin, AdvancedSearchFlarumUserMixin):
    def __init__(self):
        self.name = NAME
        self.author = AUTHOR
        self.id = ID

    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, FlarumUser, AdvancedSearchFlarumUserMixin)



from dotenv import load_dotenv
load_dotenv()

import os

from pyflarum import FlarumUser


user = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'])


if __name__ == "__main__":
    forum = user.get_forum_data()
    print(f"""
        Title: {forum.title}

        Welcome title: {forum.welcomeTitle}
        Welcome message: {forum.welcomeMessage}
    """)

    print("Groups:")
    for group in user.get_groups():
        print(f" • {group.nameSingular}")



from typing import Optional

from datetime import datetime

from .. import ExtensionMixin

from ...datetime_conversions import flarum_to_datetime
from ...flarum.core.discussions import DiscussionFromBulk, DiscussionFromNotification


AUTHOR = 'fof'
NAME = 'best-answer'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class BestAnswerDiscussionNotificationMixin(DiscussionFromNotification):
    @property
    def hasBestAnswer(self) -> bool:
        return self.attributes.get("hasBestAnswer", False)


    @property
    def bestAnswerSetAt(self) -> Optional[datetime]:
        raw = self.attributes.get("bestAnswerSetAt", None)

        return flarum_to_datetime(raw)



class BestAnswerDiscussionMixin(DiscussionFromBulk):
    @property
    def canSelectBestAnswer(self) -> bool:
        return self.attributes.get("canSelectBestAnswer", False)



class BestAnswerExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, DiscussionFromNotification, BestAnswerDiscussionNotificationMixin)
        super().mixin(self, DiscussionFromBulk, BestAnswerDiscussionMixin)



from .. import ExtensionMixin

from ...flarum.core.discussions import DiscussionFromBulk
from ...flarum.core.users import UserFromBulk


AUTHOR = 'fof'
NAME = 'byobu'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class ByobuDiscussionMixin(DiscussionFromBulk):
    @property
    def canEditRecipients(self) -> bool:
        return self.attributes.get("canEditRecipients", False)


    @property
    def canEditUserRecipients(self) -> bool:
        return self.attributes.get("canEditUserRecipients", False)


    @property
    def canEditGroupRecipients(self) -> bool:
        return self.attributes.get("canEditGroupRecipients", False)


class ByobuUserMixin(UserFromBulk):
    @property
    def blocksPd(self) -> bool:
        return self.attributes.get("blocksPd", False)


    @property
    def cannotBeDirectlyMessaged(self) -> bool:
        return self.attributes.get("cannotBeDirectlyMessaged", False)



class ByobuExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, DiscussionFromBulk, ByobuDiscussionMixin)
        super().mixin(self, UserFromBulk, ByobuUserMixin)



from dotenv import load_dotenv
load_dotenv()

import os

from pyflarum import FlarumUser


USER = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'])


if __name__ == "__main__":
    discussion = USER.get_discussions()[0] # discussion on top of the front page with default filters
    print(discussion.title)

    notifications = USER.get_notifications()
    notifications.mark_all_as_read()



from dotenv import load_dotenv
load_dotenv()

import os

from pyflarum import FlarumUser
from pyflarum.flarum.core.groups import PreparedGroup


user = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'])


if __name__ == "__main__":
    print("All groups:")
    for group in user.get_groups():
        print(f"""
            Singular: {group.nameSingular}
            Plural: {group.namePlural}
            Color: {group.color}
            Icon: {group.icon}
            Hidden: {'yes' if group.isHidden else 'no'}
        """)

    group_to_create = PreparedGroup(user=user, nameSingular="Goomba", icon="fas fa-pastafarianism")
    created = group_to_create.create()

    print(f"""Created:
        Singular: {created.nameSingular}
        Plural: {created.namePlural}
        Color: {created.color}
        Icon: {created.icon}
        Hidden: {'yes' if created.isHidden else 'no'}
    """)



from typing import TYPE_CHECKING, Iterator, Optional

if TYPE_CHECKING:
    from ...session import FlarumUser
    from ...custom_types import AnyDiscussion


from datetime import datetime

from ..core import BaseFlarumBulkObject, BaseFlarumIndividualObject
from ..core.discussions import DiscussionFromNotification
from ..core.users import UserFromBulk
from ...error_handler import parse_request
from ...datetime_conversions import flarum_to_datetime


__all__ = ['PreparedPost', 'Posts', 'Post', 'PostFromBulk', 'PostFromNotification', 'PostFromDiscussion']



class PreparedPost(BaseFlarumIndividualObject):
    """
        A prepared post that can be sent to the API.
    """

    def __init__(self, user: 'FlarumUser', discussion: Optional['AnyDiscussion']=None, content: Optional[str]=None):
        """
            ### Parameters:
            - `user` - the `pyflarum.session.FlarumUser` object that will create the post
            (see `PreparedPost.post()`).
            - `discussion` - any discussion that the post belongs to.
            - `content` - the post's content. You can use the traditional Flarum's markdown syntax.
        """

        self.user = user
        self.discussion = discussion
        self.content = content

        self.as_json = self.to_dict
        super().__init__(user=self.user, _fetched_data=self.as_json)


    @property
    def to_dict(self) -> dict:
        """
            Converts the post to a `dict`, so that
            it can be sent to the API.

            An extension might add additional data during runtime. This is the
            most basic template that Flarum requires when creating a post.
        """

        data = {
            "data": {
                "type": "posts",
                "attributes": {
                    "content": self.content
                },
                "relationships": {
                    "discussion": {
                        "data": {
                            "type": "discussions",
                            "id": self.discussion.id if self.discussion else 0
                        }
                    }
                }
            }
        }

        return data


    def post(self):
        """
            Posts/creates the post. Raises `FlarumError` on error, otherwise it returns the created `Post`.
        """

        raw = self.user.session.post(self.user.api_urls['posts'], json=self.as_json)
        json = parse_request(raw)

        return Post(user=self.user, _fetched_data=json)
    create = post



class Posts(BaseFlarumBulkObject):
    """
        A data of multiple posts fetched from `/api/posts`.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        return super().__init__(user=user, _fetched_data=_fetched_data, _listclass=PostFromBulk, _required_type='posts')


    if TYPE_CHECKING:
        def __getitem__(self, key: int) -> 'PostFromBulk': ...
        def __iter__(self) -> Iterator['PostFromBulk']: ...



class PostFromDiscussion(BaseFlarumIndividualObject):
    """
        A post from `Discussion`.
    """

    @property
    def number(self) -> Optional[int]:
        """
            The post's number/order in the discussion.
        """

        return self.attributes.get("number", None)


    @property
    def createdAt(self) -> Optional[datetime]:
        """
            The `datetime` of when was this post created.
        """

        raw = self.attributes.get("createdAt", None)

        return flarum_to_datetime(raw)


    @property
    def contentType(self) -> Optional[str]:
        """
            Post's content type. This is usually a `comment` for user-made posts,
            but this can differ if it's actually a message that a post's tags changed, or
            the discussion got moved elsewhere (these messages are treated as posts too by Flarum)
        """

        return self.attributes.get("contentType", None)


    def is_comment(self) -> bool:
        """
            Whether or not the post is comment.
        """

        return self.contentType == 'comment'


    @property
    def contentHtml(self) -> Optional[str]:
        """
            The HTML content of the post.
        """

        return self.attributes.get("contentHtml", None)


    def __restore_or_hide(self, hide: bool) -> 'Post':
        """
            Either restores or hides the post.

            This function was made to prevent code repetition - please
            use `Post.hide()` and `Post.restore()` instead.
        """

        patch_data = {
            "data": {
                "type": "posts",
                "id": self.id,
                "attributes": {
                    "isHidden": hide
                }
            }
        }

        raw = self.user.session.patch(f"{self.user.api_urls['posts']}/{self.id}", json=patch_data)
        json = parse_request(raw)

        return Post(user=self.user, _fetched_data=json)


    def hide(self) -> 'Post':
        """
            Hides the post. Raises `FlarumError` if it failed, otherwise the hidden `Post` is returned.
        """

        return self.__restore_or_hide(hide=True)


    def restore(self) -> 'Post':
        """
            Restores the post (unhides). Raises `FlarumError` if it failed, otherwise the restored `Post` is returned.
        """

        return self.__restore_or_hide(hide=False)
    unhide = restore


    def delete(self) -> True:
        """
            Removes the post forever.

            Returns `True` if the deletion was successful.
        """

        raw = self.user.session.delete(f"{self.user.api_urls['posts']}/{self.id}")
        parse_request(raw)

        return True


    def edit(self, new_post: PreparedPost) -> 'Post':
        """
            Edits the post.

            `new_post` has to be a `PreparedPost` object. Returns the `Post` after edit.
        """

        raw = self.user.session.patch(f"{self.user.api_urls['posts']}/{self.id}", json=new_post.to_dict)
        json = parse_request(raw)

        return Post(user=self.user, _fetched_data=json)



class PostFromNotification(PostFromDiscussion):
    """
        A post from `Notification`.
    """

    @property
    def content(self) -> Optional[str]:
        """
            The post's content. Doesn't contain markdown, and is just plain-text.

            Only available in `pyflarum.flarum.core.posts.PostFromNotification`.
        """

        return self.attributes.get("content", None)


    @property
    def ipAddress(self) -> Optional[str]:
        """
            The post's IP address.

            Returns `None` if you don't have the permissions to view IP address
            of the post.
        """

        return self.attributes.get("ipAddress", None)


    @property
    def editedAt(self) -> Optional[datetime]:
        """
            The `datetime` when was this post edited at.
        """

        raw = self.attributes.get("editedAt", None)

        return flarum_to_datetime(raw)


    @property
    def canEdit(self) -> bool:
        """
            Whether or not you are able to edit this post.
        """

        return self.attributes.get("canEdit", False)


    @property
    def canDelete(self) -> bool:
        """
            Whether or not you are able to delete this post.
        """

        return self.attributes.get("canDelete", False)


    @property
    def canHide(self) -> bool:
        """
            Whether or not you are able to hide this post.
        """

        return self.attributes.get("canHide", False)


    @property
    def url(self):
        """
            The post's URL.
        """

        discussion_id = self.relationships.get("discussion", {}).get("data", {}).get("id", None)

        if discussion_id:
            return f"{self.user.forum_url}/d/{discussion_id}/{self.number}"


    def get_discussion(self):
        """
            Obtains the discussion of the post.

            Returns `pyflarum.flarum.core.discussions.DiscussionFromNotification`, because its
            JSON data matches.
        """

        id_to_find = self.relationships.get("discussion", {}).get("data", {}).get("id", None)

        for possible_discussion in self._parent_included:
            if (possible_discussion.get("type", None) == "discussions") and (possible_discussion.get("id", None) == id_to_find):
                discussion = DiscussionFromNotification(user=self.user, _fetched_data=dict(data=possible_discussion, _parent_included=self._parent_included))

                return discussion

        return None


    def reply_to(self, post: PreparedPost):
        """
            Replies to this `Post` with another `PreparedPost`.
        """

        post.discussion = self.get_discussion()
        post.as_json = post.to_dict

        return post.post()


    def get_author(self):
        """
            Obtains the post's author.

            Returns `pyflarum.flarum.core.users.UserFromBulk`, because its
            JSON data is the same.
        """

        id = self.relationships.get("user", {}).get("data", {}).get("id", None)
        
        for raw_user in self._parent_included:
            if raw_user.get("id", None) == id and raw_user.get("type", None) == 'users':
                user = UserFromBulk(user=self.user, _fetched_data=dict(data=raw_user))

                return user

        return None


    def edit(self, new_data: PreparedPost):
        """
            Edits the post. The new post should be a `PreparedPost` object.
        """

        raw = self.user.session.patch(f"{self.user.api_urls['posts']}/{self.id}", json=new_data.to_dict)
        json = parse_request(raw)

        return Post(user=self.user, _fetched_data=json)



class PostFromBulk(PostFromNotification):
    """
        A post from `Posts`.
    """

    @property
    def content(self) -> NotImplemented:
        """
            This property is only available for `pyflarum.flarum.core.posts.PostFromNotification`, but
            was included here due to class inheritance.
            
            Using this will just raise `NotImplementedError`, so please do not use this.
        """

        raise NotImplementedError('The `content` property is only available for `PostFromNotification`.')



class Post(PostFromNotification):
    """
        A Flarum post.
    """

    def get_author(self):
        """
            Obtains the post's author.

            Returns `pyflarum.flarum.core.users.UserFromBulk`, because its
            JSON data is the same.
        """

        id = self.relationships.get("user", {}).get("data", {}).get("id", None)
        
        for raw_user in self.included:
            if raw_user.get("id", None) == id and raw_user.get("type", None) == 'users':
                user = UserFromBulk(user=self.user, _fetched_data=dict(data=raw_user))

                return user

        return None



from dotenv import load_dotenv
load_dotenv()

import os
#import time

from pyflarum import FlarumUser
from pyflarum.extensions.flarum import Flarum_Likes
from pyflarum.extensions import absolutely_all
from pyflarum.flarum.core.posts import Post


USER = FlarumUser(
    forum_url=os.environ['forum_url'],
    username_or_email='test',
    password=os.environ['account_password'],
    extensions=[Flarum_Likes.LikesExtension, absolutely_all.AbsolutelyAllExtension]
) # type: absolutely_all.AbsolutelyAllFlarumUserMixin


if __name__ == "__main__":
    for posts in USER.absolutely_all_posts():
        for post in posts:
            post: 'Flarum_Likes.LikesPostFromBulkMixin | Flarum_Likes.LikesPostFromDiscussionMixin'

            liked = post.like() # type: Post
            print(f"Liked post {liked.id} ({liked.url})")

            # time.sleep(1)
            continue



from dotenv import load_dotenv
load_dotenv()

from pathlib import Path

from pyflarum import FlarumUser
from pyflarum.flarum.core.filters import Filter

from pyflarum.extensions import absolutely_all
from pyflarum.extensions.flarum import Flarum_Tags


EXTENSIONS = [
    absolutely_all.AbsolutelyAllExtension,
    Flarum_Tags.TagsExtension
]

user = FlarumUser(forum_url="https://discuss.flarum.org", extensions=EXTENSIONS) # type: absolutely_all.AbsolutelyAllFlarumUserMixin


def incompatible_extensions():
    for discussions in user.absolutely_all_discussions(Filter(order_by='createdAt')):
        for discussion in discussions:
            discussion: Flarum_Tags.TagsDiscussionMixin
            tags = discussion.get_tags()

            for tag in tags:
                if tag.name == "Incompatible":
                    yield discussion


def write():
    with open(Path('tests/incompatible_extensions.txt'), 'w', encoding="utf-8") as incompatible_extensions_file:
        for discussion in incompatible_extensions():
            print(f"Incompatible: {discussion.title}")
            incompatible_extensions_file.write(f"{discussion.createdAt} | {discussion.url}\n")


if __name__ == "__main__":
    write()



from dotenv import load_dotenv
load_dotenv()

import os
from datetime import datetime

from pyflarum import FlarumUser
from pyflarum.extensions import admin
from pyflarum.extensions.flarum import Malago_Achievements

EXTENSIONS = [
    Malago_Achievements.AchievementsExtension,
    admin.AdminExtension
]


user = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'], extensions=EXTENSIONS) # type: admin.AdminFlarumUserMixin | Malago_Achievements.AchievementsAdminFlarumUserMixin


if __name__ == "__main__":
    # Update forum info:
    user.update_forum_info(
        forum_description="Peppa Pig",
        forum_title="Hi, I am Peppa Pig *oink*",
        welcome_title="This is my little brother George",
        welcome_message="This is Mommy Pig",
        user_slug_driver="And this is Daddy Pig", # surprisingly, this can be anything, even Daddy Pig
        default_route="/d/12"
    )

    # Clear custom header:
    user.update_custom_header()

    # Dynamic example: updating custom header to current time
    user.update_custom_footer(f'<div style="text-align: center; padding: 20px 0;">Extension test was last run at {datetime.now()}</div>')

    # Validation test: raise FlarumError on invalid CSS:
    user.update_custom_css('I am invalid.')


    # Enable achievements extension:
    user.enable_extension(Malago_Achievements.ID)
    all_achievements = user.get_all_achievements()

    name = "test achievement"

    # Prevent double-creation:
    can_create = True
    for achievement in all_achievements:
        print(achievement.name)

        if achievement.name == name:
            can_create = False
            break

    # Create an achievement
    if can_create:
        user.create_achievement(name="test achievements", description="owo", computation="year:1", points=100, image_url_or_fa_icon="fas fa-magic")



from .. import ExtensionMixin

from ...error_handler import parse_request
from ...flarum.core.discussions import Discussion, DiscussionFromNotification, DiscussionFromBulk


AUTHOR = 'flarum'
NAME = 'sticky'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []


class StickyDiscussionFromNotificationMixin(DiscussionFromNotification):
    def __stick_or_unstick(self, sticky: bool=True) -> Discussion:
        """
            A function to either stick or unstick the discussion, to prevent repetition.

            Use `stick()` or `unstick()` instead, please.
        """

        patch_data = {
            "data": {
                "type": "discussions",
                "id": self.id,
                "attributes": {
                    "isSticky": sticky
                }
            }
        }

        raw = self.user.session.patch(f"{self.user.api_urls['discussions']}/{self.id}", json=patch_data)
        json = parse_request(raw)

        return Discussion(user=self.user, _fetched_data=json)


    def stick(self) -> Discussion:
        """
            Stickies a discussion.
        """

        return self.__stick_or_unstick(sticky=True)


    def unstick(self) -> Discussion:
        """
            Unstickies a discussion.
        """

        return self.__stick_or_unstick(sticky=False)



class StickyDiscussionFromBulkMixin(DiscussionFromBulk, DiscussionFromNotification):
    @property
    def isSticky(self) -> bool:
        """
            Whether or not the discussion is stickied.
        """

        return self.attributes.get("isSticky", False)


    @property
    def canSticky(self) -> bool:
        """
            Whether or not you are able to stick this discussion.
        """

        return self.attributes.get("canSticky", False)



class StickyExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, DiscussionFromNotification, StickyDiscussionFromNotificationMixin)
        super().mixin(self, DiscussionFromBulk, StickyDiscussionFromBulkMixin)



from typing import Any, BinaryIO, Iterable, Optional, Literal, TYPE_CHECKING
if TYPE_CHECKING:
    from .custom_types import AnyUser


import warnings

from datetime import datetime
from requests import Session
from urllib.parse import urlparse


from .flarum.core.forum import Forum

from .flarum.core.users import MyUser, User, Users
from .flarum.core.notifications import Notifications
from .flarum.core.groups import Group, Groups

from .flarum.core.discussions import Discussion, Discussions
from .flarum.core.posts import Post, Posts

from .flarum.core.filters import Filter


from .error_handler import MissingExtensionWarning, MissingExtensionError, parse_request
from .datetime_conversions import datetime_to_flarum


from .extensions import ExtensionMixin


__all__ = ['FlarumSession', 'FlarumUser']



class FlarumSession:
    """
        The main object that carries the Flarum session.
    """

    def __init__(self, forum_url: str, username_or_email: Optional[str]=None, password: Optional[str]=None, api_endpoint: str="api", user_agent: str='pyflarum', session_object: 'Session | Any'=Session()):
        """
            ### Parameters:
            - `forum_url` - the forum URL that you want the bot to fetch/update data from.
            - `username_or_email` - optional. The username or E-mail address to log into. If `None`, then the user isn't logged in.
            - `password` - optional. The user's password. If `None`, then the user isn't logged in.
            - `api_endpoint` - the API endpoint of the forum, without slashes. This can be specified in Flarum's `config.php` and normally forums don't need to change the default `'api'`
            - `user_agent` - the user agent that will be used to make all requests. Defaults to `'pyflarum'`.
            - `session_object` - the `Session` object to make requests with. You can pass any object that supports all operations from the [requests](https://pypi.org/project/requests/) library, check [requests_cache](https://pypi.org/project/requests-cache/) as an example.
            ```
        """

        if not forum_url.startswith("http"):
            raise TypeError(f'Invalid protocol used in `forum_url` ({forum_url}) - expected either "https://" or "http://"')

        parsed_url = urlparse(forum_url, scheme="https")
        self.forum_url = f"{parsed_url.scheme}://{parsed_url.hostname}"

        self.api_endpoint = api_endpoint
        self.username_or_email = username_or_email
        self.session = session_object

        self.session.headers.update({
            "User-Agent": user_agent
        })


        self.authenticate(username_or_email, password)


    def __str__(self) -> str:
        return f'<{type(self).__name__}> authenticated={self.is_authenticated}; username={self.username_or_email}'


    def __repr__(self) -> str:
        return self.__str__()


    def authenticate(self, username_or_email: Optional[str]=None, password: Optional[str]=None) -> None:
        """
            Authenticates your user. This can be run after `FlarumUser` was initialized, to switch to a different user. You can even change
            `FlarumUser.forum_url` to login to another forum.

            ### Parameters:
            - `username_or_email` - optional. The username or E-mail address to log into. If `None`, then the user isn't logged in.
            - `password` - optional. The user's password. If `None`, then the user isn't logged in.
        """

        if username_or_email and password:
            identification_data = {"identification": username_or_email, "password": password}

            raw = self.session.post(url=f'{self.forum_url}/api/token', json=identification_data)
            token_data = parse_request(raw)

            token = token_data.get("token", None)
            user_id = token_data.get("userId", None)

            if token:
                self.session.headers.update({
                    "Authorization": f'Token {token}; userId={user_id}',
                })

                self.is_authenticated = True

            else:
                self.is_authenticated = False

        else:
            self.session.headers.update({
                "Authorization": None,
            })

            self.is_authenticated = False


    @property
    def api_urls(self) -> dict[str, str]:
        """
            Simple, hardcoded `'key: value'` `dict` of Flarum's API routes for quick access.

            API routes reference (old):
            https://github.com/flarum/flarum.github.io/blob/20322c0e6011e4f304ae7e95f41594a0b086bc27/_docs/api.md
        """

        return {
            "base": f"{self.forum_url}/{self.api_endpoint}",
            "forum": f"{self.forum_url}/{self.api_endpoint}/forum",

            "settings": f"{self.forum_url}/{self.api_endpoint}/settings",
            "extensions": f"{self.forum_url}/{self.api_endpoint}/extensions",

            "discussions": f"{self.forum_url}/{self.api_endpoint}/discussions",
            "posts": f"{self.forum_url}/{self.api_endpoint}/posts",

            "users": f"{self.forum_url}/{self.api_endpoint}/users",
            "notifications": f"{self.forum_url}/{self.api_endpoint}/notifications",
            "groups": f"{self.forum_url}/{self.api_endpoint}/groups",

            "tags": f"{self.forum_url}/{self.api_endpoint}/tags",
        }



class FlarumUser(FlarumSession, dict):
    def __init__(self, extensions: Optional[list[ExtensionMixin]]=None, **kwargs):
        """
            ### Parameters:
            - `forum_url` - the forum URL that you want the bot to fetch/update data from. This mustn't end with trailing slash (e. g.: https://domain.tld/ - wrong; https://domain.tld - correct).
            - `username_or_email` - optional. The username or E-mail address to log into. If `None`, then the bot doesn't login.
            - `password` - optional. The user's password. If `None`, then the bot doesn't login.
            - `api_endpoint` - the API endpoint of the forum, without slashes. This can be specified in Flarum's `config.php` and normally forums don't need to change the default `'api'`
            - `user_agent` - the user agent that will be used to make all requests. Defaults to `pyflarum <version>`.
            - `session_object` - the `Session` object to make requests with. You can pass any object that supports all operations from the [requests](https://pypi.org/project/requests/) library, check [requests_cache](https://pypi.org/project/requests-cache/) as an example.
            - `extensions` - a list of `ExtensionMixin` classes. These are monkey-patched on initialization. Learn more about [extensions](https://cwkevo.github.io/pyflarum/docs/#extensions).
            ```
        """

        self.extensions = extensions

        if self.extensions:
            for extension in self.extensions:
                dependencies = extension.get_dependencies(extension) # type: dict

                hard = dependencies.get("hard", None)
                soft = dependencies.get("soft", None)

                if hard and len(hard) > 0:
                    for hard_dependency in hard:
                        if hard_dependency not in self.extensions:
                            raise MissingExtensionError(f'`{extension}` hardly depends on `{hard_dependency}`. Please, include that extension too in your extension list.')

                extension.mixin(extension)

                if soft and len(soft) > 0:
                    for soft_dependency in soft:
                        if soft_dependency not in self.extensions:
                            warnings.warn(f'`{extension}` softly depends on `{soft_dependency}`. Some features might be unavailable.', MissingExtensionWarning)


        super().__init__(**kwargs)
        __json = self._fetch_user_data()


        if __json:
            self.data = MyUser(user=self, _fetched_data=dict(data=__json))
            self.username = self.data.username

        else:
            self.username = self.username_or_email


    def _fetch_user_data(self) -> Optional[dict]:
        """
            Fetches your user's JSON data.
        """

        if self.username_or_email:
            filter = Filter(limit=1)

            if '@' in self.username_or_email:
                filter.query = f'email:{self.username_or_email}'

            else:
                filter.query = self.username_or_email


            raw = self.session.get(f"{self.api_urls['users']}", params=filter.to_dict)
            json = parse_request(raw)

            for possible_user in json.get("data", [{}]):
                if '@' in self.username_or_email:
                    if possible_user.get("attributes", {}).get("email", None) == self.username_or_email:
                        return possible_user

                else:
                    if possible_user.get("attributes", {}).get("username", None) == self.username_or_email:
                        return possible_user


        return None


    def _update_user_data(self, new_data: dict) -> 'FlarumUser | User':
        """
            Updates your user data with new data, if the data belongs to you.
            Then returns updated `FlarumUser`.

            If the data doesn't belong to you, the other `User` is returned.
        """

        if int(new_data.get("data", {}).get("id", -1)) == self.data.id:
            self.data = MyUser(user=self, _fetched_data=new_data)

            return self


        else:
            user = User(user=self, _fetched_data=new_data)

            return user



    def get_forum_data(self) -> Forum:
        """
            Obtains the forum data, returns `Forum` object.
        """

        raw = self.session.get(f"{self.api_urls['base']}")
        json = parse_request(raw)


        return Forum(user=self, _fetched_data=json)


    def get_user_by_id(self, id: int) -> User:
        """
            Obtains a user by specific ID.
        """

        raw = self.session.get(f"{self.api_urls['users']}/{id}")
        json = parse_request(raw)


        return User(user=self, _fetched_data=json)


    def get_discussion_by_id(self, id: int) -> Discussion:
        """
            Obtains a discussion by specific ID.
        """

        raw = self.session.get(f"{self.api_urls['discussions']}/{id}")
        json = parse_request(raw)


        return Discussion(user=self, _fetched_data=json)


    def get_post_by_id(self, id: int) -> Post:
        """
            Obtains a post by specific ID.
        """

        raw = self.session.get(f"{self.api_urls['posts']}/{id}")
        json = parse_request(raw)


        return Post(user=self, _fetched_data=json)


    def get_discussions(self, filter: Optional[Filter]=None) -> Discussions:
        """
            Obtains all discussions from `/api/discussions`, optionally filtering results by using `filter`.
        """

        if filter:
            raw = self.session.get(f"{self.api_urls['discussions']}", params=filter.to_dict)

        else:
            raw = self.session.get(f"{self.api_urls['discussions']}")

        json = parse_request(raw)


        return Discussions(user=self, _fetched_data=json)


    def get_posts(self, filter: Optional[Filter]=None):
        """
            Obtains all posts from `/api/posts`, optionally filtering results by using `filter`.
        """

        if filter:
            raw = self.session.get(f"{self.api_urls['posts']}", params=filter.to_dict)

        else:
            raw = self.session.get(f"{self.api_urls['posts']}")

        json = parse_request(raw)


        return Posts(user=self, _fetched_data=json)


    def get_users(self, filter: Filter=None) -> Users:
        """
            Obtains all users from `/api/users`, optionally filtering results by using `filter`.
        """


        if filter:
            raw = self.session.get(f"{self.api_urls['users']}", params=filter.to_dict)

        else:
            raw = self.session.get(f"{self.api_urls['users']}")

        json = parse_request(raw)


        return Users(user=self, _fetched_data=json)


    def get_notifications(self, filter: Optional[Filter]=None) -> Notifications:
        """
            Obtains all of your notifications from `/api/notifications`, optionally filtering results by using `filter`.
        """

        if filter:
            raw = self.session.get(f"{self.api_urls['notifications']}", params=filter.to_dict)
        
        else:
            raw = self.session.get(f"{self.api_urls['notifications']}")


        json = parse_request(raw)


        return Notifications(user=self, _fetched_data=json)


    def mark_all_discussions_as_read(self, at: datetime=datetime.now()) -> Literal[True]:
        """
            Marks all discussions as read.

            Specify `at` to mark discussions as read at a specific date (strange how this is allowed, might be
            because of timezones).
        """

        post_data = {
            "data": {
                "type": "users",
                "id": self.data.id,
                "attributes": {
                    "markedAllAsReadAt": datetime_to_flarum(at)
                }
            }
        }

        raw = self.session.patch(f"{self.api_urls['users']}/{self.data.id}", json=post_data)
        parse_request(raw)


        return True


    def mark_all_notifications_as_read(self) -> Literal[True]:
        """
            Marks all notifications as read. Returns `True` when successful.
        """

        raw = self.session.post(f"{self.api_urls['notifications']}/read")
        parse_request(raw)


        return True


    def get_groups(self) -> Groups:
        """
            Obtains all groups of a forum from `/api/groups`.
        """

        raw = self.session.get(f"{self.api_urls['groups']}")
        json = parse_request(raw)


        return Groups(user=self, _fetched_data=json)


    def update_user_info(self, user: Optional['AnyUser']=None, new_username: Optional[str]=None, groups: Optional['list[str | int] | list[Group] | Groups']=None) -> 'FlarumUser | User':
        """
            Updates the info of a user (this can be your user or someone else).

            If you are updating yourself, then `FlarumUser` is returned (with the new data).
            If you are updating someone else, then the updated `User` is returned.

            ### Parameters:
            - `user` - the user to update.
            - `new_username` - the user's new username.
            - `groups` - new groups of the user. This can either be a list of `pyflarum.flarum.core.groups.Group` objects,
            or just one `pyflarum.flarum.core.groups.Groups`, or a list of `str` or `int` representing the group IDs.
        """

        id = user.id if user else self.data.id

        patch_data = {
            "data": {
                "type": "users",
                "id": id,
                "attributes": {},
                "relationships": {}
            }
        } # type: dict[str, dict[str, dict]]


        if new_username:
            patch_data['attributes']['username'] = new_username

        if groups:
            patch_data['data']['relationships'].update({"groups": {"data": [group.id if isinstance(group, Group) else group for group in groups]}})


        raw = self.session.patch(f"{self.api_urls['users']}/{id}", json=patch_data)
        json = parse_request(raw)


        return self._update_user_data(new_data=json)


    def send_password_reset_email(self) -> 'FlarumUser | User':
        """
            Allows you to send yourself a password reset E-mail.
        """

        patch_data = {"email": self.data.email}

        raw = self.session.patch(f"{self.api_urls['users']}/{self.data.id}", json=patch_data)
        json = parse_request(raw)


        return self._update_user_data(new_data=dict(data=json))


    def update_preferences(self, preferences: Iterable[tuple[str, Any]], user: Optional['AnyUser']=None) -> 'FlarumUser | User':
        """
            Updates an user's preferences.

            If no user is specified, then your user is updated.
        """

        id = user.id if user else self.data.id

        patch_data = {
            "data": {
                "type": "users",
                "id": id,
                "attributes": {
                    "preferences": {}
                }
            }
        }


        for preference, value in preferences:
            patch_data['data']['attributes']['preferences'][preference] = value


        raw = self.session.patch(f"{self.api_urls['users']}/{id}", json=patch_data)
        json = parse_request(raw)


        return self._update_user_data(new_data=json)


    def change_email(self, new_email: str, password_confirmation: str, user: Optional['AnyUser']=None) -> 'FlarumUser | User':
        """
            Changes your E-mail. If `user` is specified, then that user's E-mail is changed.

            If you are changing the E-mail of another user, you do not need to specify their password.
        """

        id = user.id if user else self.data.id

        patch_data = {
            "data": {
                "type": "users",
                "id": id,
                "attributes": {
                    "email": new_email
                }
            },
            "meta": {
                "password": password_confirmation
            }
        }

        raw = self.session.patch(f"{self.api_urls['users']}/{id}", json=patch_data)
        json = parse_request(raw)


        return self._update_user_data(new_data=json)


    def upload_user_avatar(self, file: BinaryIO, user: Optional['AnyUser']=None, file_name: str="avatar", file_type: Literal['image/png', 'image/jpeg', 'image/gif']="image/png") -> 'FlarumUser | User':
        """
            Uploads an avatar for yourself. If `user` is specified, then avatar of that user is changed.
        """

        id = user.id if user else self.data.id

        raw = self.session.post(url=f"{self.api_urls['users']}/{id}/avatar", files={ "avatar": (file_name, file, file_type) })
        json = parse_request(raw)


        return self._update_user_data(new_data=json)


    def remove_user_avatar(self, user: Optional['AnyUser']=None) -> 'FlarumUser | User':
        """
            Removes your user's avatar. If `user` is specified, then avatar of that user is removed.
        """

        raw = self.session.delete(url=f"{self.api_urls['users']}/{user.id if user else self.data.id}/avatar")
        json = parse_request(raw)
 

        return self._update_user_data(new_data=json)



from dotenv import load_dotenv
load_dotenv()

import os

from pyflarum import FlarumUser
from pyflarum.extensions import advanced_search


USER = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'], extensions=[advanced_search.AdvancedSearchExtension]) # type: advanced_search.AdvancedSearchFlarumUserMixin


if __name__ == "__main__":
    user = USER.get_user_by_username('test')

    if user:
        print(user.username)

    else:
        print("This user wasn't found.")



from dotenv import load_dotenv
load_dotenv()

import os

from pyflarum import FlarumUser
from pyflarum.extensions.flarum import Flarum_Likes
from pyflarum.flarum.core.posts import PostFromBulk
from pyflarum.flarum.core.discussions import Discussion


USER = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'], extensions=[Flarum_Likes.LikesExtension])


if __name__ == "__main__":
    discussion = USER.get_discussions()[0].get_full_data() # type: Discussion

    for post in discussion.get_posts():
        post: 'Flarum_Likes.LikesPostFromBulkMixin'
        liked_by = post.get_liked_by()

        if len(liked_by) > 0:
            print(f"Post {post.id} ({post.url}) was liked by:")
            for user in post.get_liked_by():
                print(user.username)



from dotenv import load_dotenv
load_dotenv()

from pyflarum import FlarumUser


USER = FlarumUser(forum_url="https://discuss.flarum.org")


if __name__ == "__main__":
    discussion = USER.get_discussion_by_id(28221)
    author = discussion.get_author()
    print(author.username, '=>', discussion.url)



from typing import Dict, List


class ExtensionMixin:
    """
        A base class for mixing in custom classes (extensions) into another classes.

        ### Example extension code:
        ```python

        from pyflarum.extensions import ExtensionMixin
        from pyflarum.extensions.admin import AdminExtension
        from pyflarum.session import FlarumUser


        # Lowecase:
        AUTHOR = "yourname"
        NAME = "extensionname"
        ID = f"{AUTHOR}-{NAME}"

        # List of dependencies:
        SOFT_DEPENDENCIES = [AdminExtension] # uses methods from this extension, but can run without it
        HARD_DEPENCENDIES = []



        # I recommend to use the following naming pattern: `<YourExtensionName><ClassToMixin>Mixin`
        # Example:
        class ExampleFlarumUserMixin(FlarumUser):
            @property
            def example(self):
                '''
                    Calling `FlarumUser(<...>).example` would return this.
                '''

                return "Example"


        class ExampleExtension(ExtensionMixin):
            def get_dependencies(self):
                return {
                    "soft": SOFT_DEPENDENCIES,
                    "hard": HARD_DEPENCENDIES
                }


            def mixin(self):
                super().mixin(self, FlarumUser, ExampleFlarumUserMixin)

        ```
    """

    def __init__(self):
        self.name = "Unknown"
        self.author = "Unknown"
        self.id = "N/A"


    def get_dependencies(self) -> Dict[str, List[object]]:
        """
            This should return the following `dict`:
            ```python
            {
                "hard": [<class>, <class>, ...],
                "soft": [<class>, <class>, ...]
            }
            ```

            A dependency is anything that you can pass into `FlarumUser(extensions=[...])` (e. g. an extension class).

            #### Hard-dependencies:
            - Will raise an error when they're not found in the initialized `FlarumUser` object. It is impossible for the extension
            to function without these.

            #### Soft-dependencies:
            - Will raise just a warning. It is possible for the extension to function without these, although with limitations
            (such that some functions might be unavailable).
        """

        return {
            "hard": [],
            "soft": []
        }


    def mixin(_, class_to_patch: object, class_to_mix_in: object, skip_protected: bool=True):
        """
            A function to mix-in/merge properties, methods, functions, etc... of one class into another.

            This skips all functions and properties starting with `__` (double underscore), unless `skip_protected` is False.
            
            This sets/overwrites attributes of `class_to_patch` to attributes of
            `class_to_mix_in` (monkey-patch).

            ### Example:
            ```python
            extension.mixin(myclass, pyflarum_class)
            ```
        """

        for prop, value in vars(class_to_mix_in).items():
            if prop.startswith('__') and skip_protected:
                    continue

            setattr(class_to_patch, f'{prop}', value)



from typing import Optional

from datetime import datetime

from .. import ExtensionMixin
from ...flarum.core.users import UserFromBulk


AUTHOR = 'fof'
NAME = 'username-request'
ID = f"{AUTHOR}-{NAME}"


SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class UsernameRequestUserMixin(UserFromBulk):
    @property
    def usernameHistory(self) -> Optional[list[dict[str, datetime]]]:
        raw_list = self.attributes.get("usernameHistory", None) # type: list[dict[str, int]]
        formatted_list = [] # type: list[dict[str, datetime]]

        if raw_list and isinstance(raw_list[0], dict):
            for raw_dict in raw_list:
                for username, unix in raw_dict.items():
                    dt = datetime.fromtimestamp(unix)

                    formatted_list.append({ username: dt})

            return formatted_list

        return None



class UsernameRequestExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, UserFromBulk, UsernameRequestUserMixin)



from typing import Literal, Optional, BinaryIO

from ..extensions import ExtensionMixin

from ..session import FlarumUser
from ..flarum.core import BaseFlarumIndividualObject
from ..error_handler import parse_request


AUTHOR = 'skevo'
NAME = 'admin'
ID = f"{AUTHOR}-{NAME}"


SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []


class MailSettings(BaseFlarumIndividualObject):
    @property
    def fields(self) -> dict:
        return self.attributes.get("fields", {})


    @property
    def mail(self) -> list:
        return self.fields.get("mail", [])


    @property
    def mailgun(self) -> dict:
        return self.fields.get("mailgun", {})


    @property
    def mailgun_secret(self) -> Optional[str]:
        return self.mailgun.get("mail_mailgun_secret", None)


    @property
    def mailgun_domain(self) -> Optional[str]:
        return self.mailgun.get("mail_mailgun_domain", None)


    @property
    def log(self) -> list:
        return self.fields.get("log", [])


    @property
    def smtp(self) -> dict:
        return self.fields.get("smtp", {})


    @property
    def mail_host(self) -> Optional[str]:
        return self.smtp.get("mail_host", None)


    @property
    def mail_port(self) -> Optional[str]:
        return self.smtp.get("mail_port", None)


    @property
    def mail_encryption(self) -> Optional[str]:
        return self.smtp.get("mail_encryption", None)


    @property
    def mail_username(self) -> Optional[str]:
        return self.smtp.get("mail_username", None)


    @property
    def mail_password(self) -> Optional[str]:
        return self.smtp.get("mail_password", None)


    @property
    def sending(self) -> bool:
        return self.attributes.get("sending", False)


    @property
    def errors(self) -> list:
        return self.attributes.get("errors", [])



class AdminFlarumUserMixin(FlarumUser):
    def __enable_or_disable_extension(self, id: 'str | int', enabled: bool=True) -> True:
        raw = self.session.patch(f"{self.api_urls['extensions']}/{id}", json={"enabled": enabled})
        parse_request(raw)

        return True


    def enable_extension(self, id: 'str | int') -> True:
        return self.__enable_or_disable_extension(id=id, enabled=True)


    def disable_extension(self, id: 'str | int') -> True:
        return self.__enable_or_disable_extension(id=id, enabled=False)


    def clear_cache(self) -> True:
        raw = self.session.delete(f"{self.api_urls['base']}/cache")
        parse_request(raw)

        return True


    def update_forum_info(self, forum_title: Optional[str]=None, forum_description: Optional[str]=None, default_route: Optional[str]=None, welcome_title: Optional[str]=None, welcome_message: Optional[str]=None, user_slug_driver: Optional[Literal["", "default", "id"]]=None) -> True:
        post_data = {}


        if forum_title:
            post_data["forum_title"] = forum_title

        if forum_description:
            post_data["forum_description"] = forum_description

        if default_route:
            post_data["default_route"] = default_route

        if welcome_title:
            post_data["welcome_title"] = welcome_title

        if welcome_message:
            post_data["welcome_message"] = welcome_message

        if user_slug_driver is not None:
            post_data["slug_driver_Flarum\\User\\User"] = user_slug_driver


        return self.__update_settings(post_data)


    def get_mail_settings(self) -> MailSettings:
        raw = self.session.get(f"{self.api_urls['base']}/mail/settings")
        json = parse_request(raw)

        return MailSettings(json)


    def update_mail_settings(self, mail_from: Optional[str]=None, mail_encryption: Optional[str]=None, mail_username: Optional[str]=None, mail_password: Optional[str]=None):
        post_data = {}


        if mail_from:
            post_data["mail_from"] = mail_from

        if mail_encryption:
            post_data["mail_encryption"] = mail_encryption

        if mail_username:
            post_data["mail_username"] = mail_username

        if mail_password:
            post_data["mail_password"] = mail_password


        return self.__update_settings(post_data)


    def send_test_mail(self) -> True:
        raw = self.session.post(f"{self.api_urls['base']}/mail/test")
        parse_request(raw)

        return True


    def update_appearance(self, theme_primary_color: Optional[str]=None, theme_secondary_color: Optional[str]=None, theme_dark_mode: Optional[bool]=None, theme_colored_header: bool=None) -> True:
        post_data = {}


        if theme_primary_color:
            post_data["theme_primary_color"] = theme_primary_color

        if theme_secondary_color:
            post_data["theme_secondary_color"] = theme_secondary_color

        if isinstance(theme_dark_mode, bool):
            post_data["theme_dark_mode"] = theme_dark_mode

        if isinstance(theme_colored_header, bool):
            post_data["theme_colored_header"] = theme_colored_header


        raw = self.session.post(f"{self.api_urls['base']}/settings", json=post_data)
        parse_request(raw)

        return True


    def upload_logo(self, file: BinaryIO, file_name: str="logo", file_type: Literal['image/png', 'image/jpeg', 'image/gif']="image/png") -> True:
        raw = self.session.post(url=f"{self.api_urls['base']}/logo", files={ "logo": (file_name, file, file_type) })

        parse_request(raw)

        return True


    def remove_logo(self) -> True:
        raw = self.session.delete(f"{self.api_urls['base']}/logo")
        parse_request(raw)

        return True


    def upload_favicon(self, file: BinaryIO, file_name: str="favicon", file_type: Literal['image/png', 'image/jpeg', 'image/gif']="image/png") -> True:
        raw = self.session.post(url=f"{self.api_urls['base']}/favicon", files={ "favicon": (file_name, file, file_type) })

        parse_request(raw)

        return True


    def remove_favicon(self) -> True:
        raw = self.session.delete(f"{self.api_urls['base']}/favicon")
        parse_request(raw)

        return True


    def update_custom_header(self, header: Optional[str]=None) -> True:
        return self.__update_settings({ "custom_header": header if header else "" })


    def update_custom_footer(self, footer: Optional[str]=None) -> True:
        return self.__update_settings({ "custom_footer": footer if footer else "" })


    def update_custom_css(self, css: Optional[str]=None) -> True:
        return self.__update_settings({ "custom_header": css if css else "" })


    def __update_settings(self, data: dict) -> True:
        raw = self.session.post(f"{self.api_urls['base']}/settings", json=data)
        parse_request(raw)

        return True



class AdminExtension(ExtensionMixin, AdminFlarumUserMixin):
    def __init__(self):
        self.name = NAME
        self.author = AUTHOR
        self.id = ID

    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, FlarumUser, AdminFlarumUserMixin)



from dotenv import load_dotenv
load_dotenv()

import os
import requests

from pyflarum import FlarumUser
from pyflarum.extensions.flarum import FoF_UserBio


user = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'], extensions=[FoF_UserBio.UserBioExtension]) # type: FoF_UserBio.UserBioFlarumUserMixin


if __name__ == "__main__":
    # Get random quote:
    quote = requests.get("http://api.quotable.io/random").json() # type: dict

    # Set it as your bio:
    updated = user.update_user_bio(f"{quote['content']}\n\n- {quote.get('author', 'Unknown')}") # type: FlarumUser
    print(updated.data.bio) # can't type hint updated.data



from dotenv import load_dotenv
load_dotenv()

#import time

from pyflarum import FlarumUser
from pyflarum.extensions.flarum import Flarum_Likes
from pyflarum.extensions import absolutely_all


USER = FlarumUser(
    forum_url="https://discuss.flarum.org",
    extensions=[Flarum_Likes.LikesExtension, absolutely_all.AbsolutelyAllExtension]
) # type: absolutely_all.AbsolutelyAllFlarumUserMixin


if __name__ == "__main__":
    discussion = USER.get_discussion_by_id(7585)

    for posts in USER.get_all_posts_from_discussion(discussion):
        for post in posts:
            print(post.url)
        print('-' * 20)



from typing import TYPE_CHECKING, Iterator, Optional

# Avoid my greatest enemy in Python: circular import:
if TYPE_CHECKING:
    from ...session import FlarumUser

from ...error_handler import parse_request
from ..core import BaseFlarumBulkObject, BaseFlarumIndividualObject


__all__ = ['PreparedGroup', 'Groups', 'Group']



class PreparedGroup(BaseFlarumIndividualObject):
    def __init__(self, user: 'FlarumUser', nameSingular: str, namePlural: Optional[str]=None, color: Optional[str]=None, icon: Optional[str]=None, isHidden: bool=False):
        self.user = user

        self.nameSingular = nameSingular
        self.namePlural = (namePlural if namePlural else self.nameSingular)
        self.color = color
        self.icon = icon
        self.isHidden = isHidden

        self.as_json = self.to_dict
        super().__init__(user=self.user, _fetched_data=self.as_json)


    @property
    def to_dict(self) -> dict:
        """
            Converts the group to a `dict`, so that
            it can be sent to the API.

            An extension might add additional data during runtime. This is the
            most basic template that Flarum requires when creating a group.
        """

        data = {
            "data": {
                "type": "groups",
                "attributes": {
                    "nameSingular": self.nameSingular,
                    "namePlural": self.namePlural,
                    "color": self.color,
                    "icon": self.icon,
                    "isHidden": self.isHidden
                }
            }
        }

        return data


    def create(self) -> 'Group':
        """
            Creates the group. Returns the created `Group`.
        """

        raw = self.user.session.post(self.user.api_urls['groups'], json=self.as_json)
        json = parse_request(raw)

        return Group(user=self.user, _fetched_data=json)



class Groups(BaseFlarumBulkObject):
    """
        A data of multiple groups fetched from the API.
    """


    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        return super().__init__(user=user, _fetched_data=_fetched_data, _listclass=Group, _required_type='groups')


    if TYPE_CHECKING:
        def __getitem__(self, key: int) -> 'Group': ...
        def __iter__(self) -> Iterator['Group']: ...



class Group(BaseFlarumIndividualObject):
    """
        A Flarum group.
    """


    @property
    def nameSingular(self) -> Optional[str]:
        """
            Singular form of the group's name.
        """

        return self.attributes.get("nameSingular", None)


    @property
    def namePlural(self) -> Optional[str]:
        """
            Plural form of the group's name.
        """

        return self.attributes.get("namePlural", None)


    @property
    def color(self) -> Optional[str]:
        """
            The color of the group.
        """

        return self.attributes.get("color", None)


    @property
    def icon(self) -> Optional[str]:
        """
            [FontAwesome](https://fontawesome.com/v5.15/icons?d=gallery) icon of the group.
        """

        return self.attributes.get("icon", None)


    @property
    def isHidden(self) -> bool:
        """
            Whether or not the group is hidden on the forum.
        """

        return bool(self.attributes.get("isHidden", 0))


    def edit(self, new_data: PreparedGroup) -> 'Group':
        """
            Edits the group with new `PreparedGroup`.

            Returns the edited `Group`
        """

        raw = self.user.session.patch(f"{self.user.api_urls['groups']}/{self.id}", json=new_data.to_dict)
        json = parse_request(raw)

        return Group(user=self.user, _fetched_data=json)


    def delete(self) -> True:
        """
            Removes the group forever. This is irreversible!

            Returns `True` when the deletion was successful.
        """

        raw = self.user.session.delete(f"{self.user.api_urls['groups']}/{self.id}")
        parse_request(raw)

        return True




from dotenv import load_dotenv
load_dotenv()

import os
import requests

from pyflarum import FlarumUser
from pyflarum.extensions import admin


user = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'], extensions=[admin.AdminExtension]) # type: admin.AdminFlarumUserMixin


if __name__ == "__main__":
    # Get random birb image:
    birb_link = requests.get("https://some-random-api.ml/img/birb").json()['link']
    birb_image = requests.get(birb_link, stream=True).content

    # Upload it as avatar:
    updated_user = user.upload_user_avatar(birb_image)
    print(updated_user.data.avatarUrl)

    # ...or for someone else:
    _user = user.get_user_by_id(3)
    user.upload_user_avatar(birb_image, user=_user)
    print(user.data.id)

    # Also, why not upload it as logo?
    user.upload_logo(birb_image)

    # Remove uploaded favicon:
    user.remove_favicon()
    print("OK")



from .session import FlarumUser
from .error_handler import FlarumError

from .flarum.core.discussions import PreparedDiscussion, Discussions, Discussion, DiscussionFromBulk, DiscussionFromNotification
from .flarum.core.posts import PreparedPost, Posts, Post, PostFromBulk, PostFromNotification, PostFromDiscussion

from .flarum.core.users import Users, User, UserFromBulk, UserFromNotification, MyUser
from .flarum.core.groups import PreparedGroup, Groups, Group

from .flarum.core.filters import Filter
from .flarum.core.notifications import Notifications, Notification
from .flarum.core.forum import Forum


__all__ = [
    'FlarumUser', 'FlarumError',

    'PreparedDiscussion', 'Discussions', 'Discussion', 'DiscussionFromBulk', 'DiscussionFromNotification',
    'PreparedPost', 'Posts', 'Post', 'PostFromBulk', 'PostFromNotification', 'PostFromDiscussion',

    'Users', 'User', 'UserFromBulk', 'UserFromNotification', 'MyUser',
    'PreparedGroup', 'Groups', 'Group',

    'Filter',
    'Notifications', 'Notification',
    'Forum'
]



from dotenv import load_dotenv
load_dotenv()

import os

from pyflarum import FlarumUser

from pyflarum.flarum.core.filters import Filter

from pyflarum.extensions import absolutely_all
from pyflarum.extensions.flarum import Flarum_Subscriptions


EXTENSIONS = [
    absolutely_all.AbsolutelyAllExtension,
    Flarum_Subscriptions.SubscriptionsExtension
]

user = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'], extensions=EXTENSIONS) # type: absolutely_all.AbsolutelyAllFlarumUserMixin


def follow_all():
    for discussions in user.absolutely_all_discussions(Filter(query='is:ignored')):
        for discussion in discussions:
            discussion: Flarum_Subscriptions.SubscriptionsDiscussionFromBulkMixin
            discussion = discussion.follow()
            print(f"Now following ignored discussion {discussion.id} ({discussion.url})")


if __name__ == "__main__":
    follow_all()



from .. import ExtensionMixin

from ...flarum.core.discussions import DiscussionFromBulk


AUTHOR = 'blomstra'
NAME = 'realtime'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class RealtimeDiscussionMixin(DiscussionFromBulk):
    @property
    def canViewWhoTypes(self) -> bool:
        """
            Whether or not you can view who is typing in real time.
        """

        return self.attributes.get("canViewWhoTypes", False)



class RealtimeExtension(ExtensionMixin):
    """
        https://extiverse.com/extension/blomstra/realtime
    """

    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, DiscussionFromBulk, RealtimeDiscussionMixin)



from pathlib import Path

import sys
sys.path.append('.')

import setuptools


__description__ = "An unofficial Python package for manipulating with Flarum's API"
__author__      = "SKevo"
__copyright__   = "Copyright 2021, SKevo"
__credits__     = ["SKevo"]
__license__     = "GPLv3"
__version__     = "v1.0.9.1-beta"
__maintainer__  = "SKevo"
__email__       = "me@kevo.link"
__status__      = "4 - Beta"


# Overwrite docstring, so pdoc can render it:
try:
    with open(f"{Path(__file__).parent.absolute()}{Path('/README.md')}", 'r', encoding="UTF-8") as readme:
        __readme__ = readme.read()

except:
    __readme__ = "Failed to read README.md!\n\n(c) SKevo" # type: str


__doc__ = __readme__


setuptools.setup(
    name = 'pyFlarum',
    packages = setuptools.find_packages(),

    long_description=__readme__,
    long_description_content_type='text/markdown',

    version = __version__,
    license=__license__,
    description = __description__,
    keywords = ['Flarum', 'forum software', 'api', 'Flarum api', 'bot', 'userbot', 'database', 'flarum database', 'server', 'client'],

    author = __author__,
    author_email = __email__,

    url = 'https://github.com/CWKevo/pyflarum',

    install_requires=[
        'requests'
    ],

    classifiers=[
        f'Development Status :: {__status__}',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



from dotenv import load_dotenv
load_dotenv()

import os

from pyflarum import FlarumUser

from pyflarum.flarum.core.posts import PostFromNotification

from pyflarum.extensions import absolutely_all


EXTENSIONS = [
    absolutely_all.AbsolutelyAllExtension
]

user = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'], extensions=EXTENSIONS) # type: absolutely_all.AbsolutelyAllFlarumUserMixin


def all_notifications():
    for notifications in user.absolutely_all_notifications():
        for notification in notifications:
            print(notification.id)

            if not notification.isRead:
                subject = notification.get_subject()

                if isinstance(subject, PostFromNotification):
                    print(subject.content)



if __name__ == "__main__":
    print(f"{user.username}'s notification IDs:")
    all_notifications()



from dotenv import load_dotenv
load_dotenv()

import os

from datetime import datetime
from pathlib import Path

from requests_cache.session import CachedSession

from pyflarum import FlarumUser
from pyflarum.flarum.core.filters import Filter

from pyflarum.extensions import absolutely_all

EXTENSIONS = [
    absolutely_all.AbsolutelyAllExtension
]


def scrap_all(file_name: 'str | bytes | Path'="scrapped.html"):
    USER = FlarumUser(forum_url=os.environ['forum_url'], extensions=EXTENSIONS, session_object=CachedSession()) # type: absolutely_all.AbsolutelyAllFlarumUserMixin

    with open(file_name, 'w', encoding='UTF-8') as scrap:
        header = f"""<html><head><title>All discussions from {USER.forum_url}</title></head><body><div style="margin: 2rem 0; text-align: center; color: blue;"><h1>All discussions from <a href="{USER.forum_url}">{USER.forum_url}</a> ({datetime.now().strftime(r'%H:%M:%S %d. %m. %Y')})</h1></div><div id="data">\n"""

        scrap.write(header)

        for discussions in USER.absolutely_all_discussions(Filter(order_by="createdAt")):
            try:
                for _discussion in discussions:
                        #time.sleep(5) # prevent 429
                        discussion = _discussion.get_full_data()

                        print(f"Writing data for {discussion.url}...")

                        data = f"""<div id="d-{discussion.id}" style="margin: 2rem; background-color: lavender; border: 3px double skyblue; padding: 1rem; border-radius: 2rem;"><h1><a href="{discussion.url}">Discussion #{discussion.id}</a></h1><div id="d-{discussion.id}-posts">\n"""
                        scrap.write(data)

                        for posts in USER.get_all_posts_from_discussion(discussion):
                            for post in posts:
                                if post.contentHtml:
                                    data = f"""<div id="d-{discussion.id}-{post.number}"><h3><a href="{post.url}">Post #{post.number} in discussion #{discussion.id}</a></h3><p><i>By: {post.get_author().username} @ {post.createdAt}</i></p><div>{post.contentHtml}</div><hr/><br/></div>\n"""
                                    scrap.write(data)

                        scrap.write("""</div></div>""")

            except KeyboardInterrupt:
                break

        footer = """</div><p style="text-align: center; font-size: 2rem;">The end.</p></body></html>\n"""

        scrap.write(footer)
        print("Done!")



if __name__ == "__main__":
    from tests import ROOT_PATH

    scrap_all(f"{ROOT_PATH}/tests/scrapped.html")



from typing import Literal, Optional, Type

from .. import ExtensionMixin

from ...flarum.core.discussions import Discussion, DiscussionFromNotification, DiscussionFromBulk

from ...error_handler import parse_request


AUTHOR = 'flarum'
NAME = 'subscriptions'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []


class SubscriptionsDiscussionFromNotificationMixin(DiscussionFromNotification):
    def __change_subscription_state(self, state: Optional[Literal['following', 'ignoring']]):
        """
            Changes the subscription state of a discussion.

            This function is intended to prevent code repetition - you are
            supposed to be using `follow()`, `unfollow()` or `ignore()`.
        """

        post_data = {
            "data": {
                "type": "discussions",
                "id": self.id,
                "attributes": {
                    "subscription": state
                }
            }
        }

        raw = self.user.session.patch(f"{self.user.api_urls['discussions']}/{self.id}", json=post_data)
        json = parse_request(raw)

        return Discussion(user=self.user, _fetched_data=json)
    

    def follow(self):
        """
            Follow the discussion and be notified of all new activity.
        """

        return self.__change_subscription_state(state='following')


    def unfollow(self):
        """
            Unfollow the discussion, but be notified when someone mentions you.
        """

        return self.__change_subscription_state(state=None)


    def ignore(self):
        """
            Ignore the discussion, never be mentioned.

            Note that this will also hide the discussion from `Discussions`.
            Currently, the only ways to access ignored Flarum discussions that I am aware of are:
            1. Accessing the discussion directly (by ID).
            2. Using `pyflarum.flarum.core.filters.Filter` (e. g. `Filter(query="is:ignored")`).
        """
        return self.__change_subscription_state(state='ignore')



class SubscriptionsDiscussionFromBulkMixin(DiscussionFromBulk, SubscriptionsDiscussionFromNotificationMixin):
    @property
    def subscription(self) -> Optional[Literal['following', 'ignoring']]:
        """
            Get the current subscription state of the discussion.
        """

        return self.attributes.get("subscription", None)



class SubscriptionsExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, DiscussionFromNotification, SubscriptionsDiscussionFromNotificationMixin)
        super().mixin(self, DiscussionFromBulk, SubscriptionsDiscussionFromBulkMixin)



from .. import ExtensionMixin

from ...flarum.core.discussions import DiscussionFromBulk


AUTHOR = 'askvortsov'
NAME = 'reply-templates'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class ReplyTemplatesDiscussionFromBulkMixin(DiscussionFromBulk):
    @property
    def replyTemplate(self) -> str:
        """
            The reply template for the discussion.
        """

        return self.attributes.get("replyTemplate", "")


    @property
    def canManageReplyTemplates(self) -> bool:
        """
            Whether or not you are able to manage the discussion's reply templates.
        """

        return self.attributes.get("canManageReplyTemplates", False)



class ReplyTemplatesExtension(ExtensionMixin):
    """
        https://extiverse.com/extension/askvortsov/flarum-discussion-templates
    """

    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, DiscussionFromBulk, ReplyTemplatesDiscussionFromBulkMixin)



from .. import ExtensionMixin

from ...error_handler import parse_request
from ...flarum.core.discussions import Discussion, DiscussionFromBulk, DiscussionFromNotification


AUTHOR = 'flarum'
NAME = 'lock'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class LockDiscussionFromNotificationMixin(DiscussionFromNotification):
    def __lock_or_unlock(self, locked: bool=True) -> Discussion:
        """
            A function to either lock or unlock the post, to prevent repetition.

            Use `lock()` or `unlock()` instead, please.
        """

        patch_data = {
            "data": {
                "type": "discussions",
                "id": self.id,
                "attributes": {
                    "isLocked": locked
                }
            }
        }

        raw = self.user.session.patch(f"{self.user.api_urls['discussions']}/{self.id}", json=patch_data)
        json = parse_request(raw)

        return Discussion(user=self.user, _fetched_data=json)


    def lock(self) -> Discussion:
        """
            Locks the discussion.
        """

        return self.__lock_or_unlock(locked=True)


    def unlock(self) -> Discussion:
        """
            Unlocks the discussion.
        """

        return self.__lock_or_unlock(locked=False)



class LockDiscussionFromBulkMixin(DiscussionFromBulk, LockDiscussionFromNotificationMixin):
    @property
    def isLocked(self) -> bool:
        """
            Whether or not the discussion is locked.
        """

        return self.attributes.get("isLocked", False)


    @property
    def canLock(self) -> bool:
        """
            Whether or not you are able to lock the discussion.
        """

        return self.attributes.get("canLock", False)



class LockExtension(ExtensionMixin):
    """
        https://packagist.org/packages/flarum/lock
    """

    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, DiscussionFromNotification, LockDiscussionFromNotificationMixin)
        super().mixin(self, DiscussionFromBulk, LockDiscussionFromBulkMixin)



from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from ...custom_types import AnyUser
    from ...flarum.core.users import User

from .. import ExtensionMixin
from ...session import FlarumUser

from ...flarum.core.forum import Forum
from ...flarum.core.users import UserFromBulk
from ...error_handler import parse_request


AUTHOR = 'fof'
NAME = 'user-bio'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class UserBioForumMixin(Forum):
    @property
    def max_bio_length(self) -> bool:
        return self.attributes.get("fof-user-bio.maxLength", False)



class UserBioFlarumUserMixin(FlarumUser):
    def update_user_bio(self, bio: Optional[str]=None, user: Optional['AnyUser']=None) -> 'FlarumUser | User':
        id = user.id if user else self.data.id

        post_data = {
            "data": {
                "type": "users",
                "id": self.data.id,
                "attributes": {
                    "bio": bio if bio else ""
                }
            }
        }

        raw = self.session.patch(f"{self.api_urls['users']}/{id}", json=post_data)
        json = parse_request(raw)


        return self._update_user_data(new_data=json)



class UserBioUserFromBulkMixin(UserFromBulk):
    @property
    def bio(self) -> Optional[str]:
        return self.attributes.get("bio", None)


    @property
    def canViewBio(self) -> bool:
        return self.attributes.get("canViewBio", False)


    @property
    def canEditBio(self) -> bool:
        return self.attributes.get("canEditBio", False)


class UserBioExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, FlarumUser, UserBioFlarumUserMixin)
        super().mixin(self, UserFromBulk, UserBioUserFromBulkMixin)
        super().mixin(self, Forum, UserBioForumMixin)



from .. import ExtensionMixin

from ...flarum.core.forum import Forum


AUTHOR = "flarum"
NAME = "markdown"
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class ForumMixin(Forum):
    @property
    def markdown_mdarea(self) -> bool:
        """
            Whether or not the MDArea is enabled for markdown.
        """

        return self.attributes.get("flarum-markdown.mdarea", False)



class ExampleExtension(ExtensionMixin):
    """
        https://packagist.org/packages/flarum/markdown
    """

    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, Forum, ForumMixin)



from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from ...session import FlarumUser
    from ...custom_types import AnyFlarumClass


__all__ = ['BaseFlarumObject', 'BaseFlarumBulkObject', 'BaseFlarumIndividualObject']



class BaseFlarumObject(dict):
    """
        The base Flarum object - all API objects have properties of this object.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        """
            ### Parameters:
            - `user` - the `pyflarum.session.FlarumUser` object, required to make additional API calls.
            - `_fetched_data` - the JSON data that was fetched from the API.

            I strongly discourage from forging objects this way, unless you are creating an extension.
        """

        self.user = user
        super().__init__(_fetched_data)


    @property
    def data(self) -> 'list[dict] | dict':
        """
            A raw `dict` of the object's data.
        """

        return self.get("data", {})



class BaseFlarumBulkObject(list):
    """
        The base object for Flarum "bulk" objects - all API objects
        that contain other objects have these properties.

        Examples of bulk objects:
        - `pyflarum.flarum.core.users.UserFromBulk`
        - `pyflarum.flarum.core.discussions.DiscussionFromBulk`
        - `pyflarum.flarum.core.posts.PostFromBulk`
    """


    def __init__(self, user: 'FlarumUser', _fetched_data: dict, _listclass: 'AnyFlarumClass', _required_type: str):
        self.links = _fetched_data.get("links", {}) # type: dict
        self.included = _fetched_data.get("included", [{}]) # type: list[dict]

        self.user = user
        converted = [_listclass(user=user, _fetched_data=dict(data=data, _parent_included=self.included)) for data in _fetched_data.get("data", []) if data.get("type", _required_type) == _required_type]
        super().__init__(converted)


    @property
    def first_link(self) -> Optional[str]:
        """
            First link in the API.
        """

        return self.links.get("first", None)


    @property
    def previous_link(self) -> Optional[str]:
        """
            Previous link in the API.
        """

        return self.links.get("prev", None)


    @property
    def next_link(self) -> Optional[str]:
        """
            Next link in the API.
        """

        return self.links.get("next", None)


    if TYPE_CHECKING:
        @property
        def included(self) -> list[dict]:
            """
                Returns raw list of JSON included data.

                Learn more about included data [here](https://cwkevo.github.io/pyflarum/docs/#included-data)
            """
            ...


        @property
        def links(self) -> dict:
            """
                Returns raw list of API links.
            """
            ...



class BaseFlarumIndividualObject(BaseFlarumObject):
    """
        Base object for Flarum "individual" objects - all
        objects have these properties.

        Examples of "individual" objects:
        - `pyflarum.flarum.core.discussions.Discussion`
        - `pyflarum.flarum.core.posts.Post`
        - `pyflarum.flarum.core.PostFromDiscussion`
    """

    @property
    def type(self) -> Optional[str]:
        """
            The type of the object.
            
            This should always be the plural form of Flarum's name of the object, e. g. `discussions`, `posts`, `users`, etc...
        """

        return self.data.get("type", None)


    @property
    def id(self) -> Optional[int]:
        """
            The `int` ID of the object. This should always be unique for the object's type.
        """

        raw = self.data.get("id", None)

        if raw:
            return int(raw)


    @property
    def attributes(self) -> dict:
        """
            Raw `dict` of the object's attributes.
        """

        return self.data.get("attributes", {})


    @property
    def relationships(self) -> dict:
        """
            Raw `dict` of the object's relationships with
            other objects.

            This contains references to objects in the included data.
            Read more about [included data](https://cwkevo.github.io/pyflarum/docs/#included-data).
        """

        return self.data.get("relationships", {})


    @property
    def included(self) -> list[dict]:
        """
            Returns raw list of JSON included data.

            Learn more about included data [here](https://cwkevo.github.io/pyflarum/docs/#included-data).

            Note: Not all "individual" objects have this property, usually the low-level ones (such as `PostFromX`, `DiscussionFromX`, etc...).
        """

        return self.get("included", {})


    @property
    def _parent_included(self) -> list[dict]:
        """
            Raw data of the parent's included JSON data.

            Returns empty `list` of empty `dict` if there is no parent data.

            Read more about [_parent_included](https://cwkevo.github.io/pyflarum/docs/#parent-included).
        """

        return self.get("_parent_included", [{}])



from datetime import timedelta

from dotenv import load_dotenv
load_dotenv()

import os

from pyflarum import FlarumUser

from pyflarum.extensions import admin
from pyflarum.extensions.flarum import FoF_Spamblock
from pyflarum.extensions.flarum import Flarum_Suspend


EXTENSIONS = [
    admin.AdminExtension,
    FoF_Spamblock.SpamblockExtension,
    Flarum_Suspend.SuspendExtension
]


user = FlarumUser(forum_url=os.environ['forum_url'], username_or_email='test', password=os.environ['account_password'], extensions=EXTENSIONS)


if __name__ == "__main__":
    id_to_block = 3

    to_block = user.get_user_by_id(id_to_block) # type: FoF_Spamblock.SpamblockUserMixin | Flarum_Suspend.SuspendUserMixin
    blocked = to_block.spamblock()
    print(f"Success: {blocked}")

    to_block.suspend(suspended_for=timedelta(minutes=15))
    print("Suspended for 15 minutes instead.")
